import os
import uuid
import time
import warnings
from .core_sina import KoshSinaObject, kosh_pickler
from .utils import get_graph
from .utils import compute_fast_sha
from .utils import compute_long_sha
from .utils import cleanup_sina_record_from_kosh_sync
from .utils import update_json_file_with_records_and_relationships
from .utils import __check_valid_connection_type__
from . import lock_strategies
import kosh
import six
try:
    import orjson
except ImportError:
    import json as orjson  # noqa
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


class KoshDataset(KoshSinaObject):
    def __init__(self, id, store, schema=None, record=None, kosh_type=None):
        """KoshSinaDataset Sina representation of Kosh Dataset

        :param id: dataset's unique Id
        :type id: str
        :param store: store containing the dataset
        :type store: KoshSinaStore
        :param schema: Kosh schema validator
        :type schema: KoshSchema
        :param record: to avoid looking up in sina pass sina record
        :type record: Record
        :param kosh_type: type of Kosh object (dataset, file, project, ...)
        :type kosh_type: str
        """
        with store.lock_strategy:
            if kosh_type is None:
                kosh_type = store._dataset_record_type
            super(KoshDataset, self).__init__(id, kosh_type=kosh_type,
                                              protected=[
                                                "__name__", "__creator__", "__store__",
                                                "_associated_data_", "__features__"],
                                              record_handler=store.__record_handler__,
                                              store=store, schema=schema, record=record)
            self.__dict__["__record_handler__"] = store.__record_handler__
            if record is None:
                record = self.get_record()
            try:
                self.__dict__["__creator__"] = record["data"]["creator"]["value"]
            except Exception:
                pass
            try:
                self.__dict__["__name__"] = record["data"]["name"]["value"]
            except Exception:
                pass
            if schema is not None or "schema" in record["data"]:
                self.validate()

    @lock_strategies.lock_method
    def __str__(self):
        """string representation"""
        import reprlib
        if self.__store__.verbose_attributes:
            def reprtool(item):
                return item
        else:
            def reprtool(item):
                if isinstance(item, str):
                    return reprlib.repr(item)[1:-1]
                else:
                    return reprlib.repr(item)
        st = ""
        st += "KOSH DATASET\n"
        st += "\tid: {}\n".format(self.id)
        try:
            st += "\tname: {}\n".format(self.__name__)
        except Exception:
            st += "\tname: ???\n"
        try:
            st += "\tcreator: {}\n".format(self.creator)
        except Exception:
            st += "\tcreator: ???\n"
        atts = self.__attributes__
        if len(atts) > 0:
            st += "\n--- Attributes ---\n"
            for a in sorted(atts):
                if a == "_associated_data_" or "_ENSEMBLE_TAG_" in a:  # Remove associated data and ensemble tags
                    continue
                if not self.is_ensemble_attribute(a):
                    st += "\t{}: {}\n".format(a, reprtool(atts[a]))
        if self._associated_data_ is not None:
            st += "--- Associated Data ({})---\n".format(
                len(self._associated_data_))
            # Let's organize per mime_type
            associated = {}
            for a in self._associated_data_:
                if a == self.id:
                    st2 = "internal ( {} )".format(
                        ", ".join(self.get_record()["curve_sets"].keys()))
                    if "sina/curve" not in associated:
                        associated["sina/curve"] = [st2, ]
                    else:
                        associated["sina/curve"].append(st2)
                else:
                    if "__uri__" in a:
                        a_id, a_uri = a.split("__uri__")
                        a_mime_type = self.get_record(
                        )["files"][a_uri]["mimetype"]
                    else:
                        a_id, a_uri = a, None
                    a_obj = self.__store__._load(a_id)
                    if a_uri is None:
                        a_uri = a_obj.uri
                        a_mime_type = a_obj.mime_type
                    st2 = "{a_uri} ( {a} )".format(a_uri=a_uri, a=a_id)
                    if a_mime_type not in associated:
                        associated[a_mime_type] = [st2, ]
                    else:
                        associated[a_mime_type].append(st2)
            for mime in sorted(associated):
                st += "\tMime_type: {mime}".format(mime=mime)
                for uri in sorted(associated[mime]):
                    st += "\n\t\t{uri}".format(uri=uri)
                st += "\n"
        ensembles = tuple(self.get_ensembles())
        st += "--- Ensembles ({})---".format(len(ensembles))
        st += "\n\t" + str([str(x.id) for x in ensembles])
        st += "\n--- Ensemble Attributes ---\n"
        for ensemble in ensembles:
            st += "\t--- Ensemble {} ---\n".format(ensemble.id)
            eas = ensemble.list_attributes()
            for ignore in ['creator', 'id', 'name']:
                eas.remove(ignore)
            eas.sort()
            st += f"\t\t{reprtool(eas)}\n"

            ensemble_tags = self.list_ensemble_tags(ensemble.id)
            ensemble_tags.sort()
            if ensemble_tags:
                ensemble_tags = [et.replace(f"{ensemble.id}_ENSEMBLE_TAG_", "") for et in ensemble_tags]
                st += "\t\t--- Ensemble Tags ---\n"
                st += f"\t\t\t{ensemble_tags}\n"
        if self.alias_feature is not {}:
            st += '--- Alias Feature Dictionary ---'
            for key, val in self.alias_feature.items():
                st += f"\n\t{key}: {val}"
        return st

    @lock_strategies.lock_method
    def _repr_pretty_(self, p, cycle):
        """Pretty display in Ipython"""
        p.text(self.__str__())

    @lock_strategies.lock_method
    def cleanup_files(self, dry_run=False, interactive=False,
                      clean_fastsha=False, **search_keys):
        """Cleanup the dataset from references to dead files
        Also updates the fast_shas if necessary
        You can filter associated objects by passing key=values
        e.g mime_type=hdf5 will only dissociate non-existing files associated with mime_type hdf5
        some_att=some_val will only dissociate non-existing files associated and having the attribute
        'some_att' with value of 'some_val'
        returns list of uris to be removed.
        :param dry_run: Only does a dry_run
        :type dry_run: bool
        :param interactive: interactive mode, ask before dissociating
        :type interactive: bool
        :param clean_fastsha: Do we want to update fast_sha if it changed?
        :type clean_fastsha: bool
        :returns: list of uris (to be) removed or updated
        :rtype: list
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        bads = []
        for associated in self.find(**search_keys):
            clean = 'n'
            if not os.path.exists(associated.uri):  # Ok this is gone
                bads.append(associated.uri)
                if dry_run:  # Dry run
                    clean = 'n'
                elif interactive:
                    clean = input("\tDo you want to dissociate {} (mime_type: {})? [Y/n]".format(
                        associated.uri, associated.mime_type)).strip()
                    if len(clean) > 0:
                        clean = clean[0]
                        clean = clean.lower()
                    else:
                        clean = 'y'
                else:
                    clean = 'y'
                if clean == 'y':
                    self.dissociate(associated.uri)
            elif clean_fastsha:
                # file still exists
                # We might want to update its fast sha
                fast_sha = compute_fast_sha(associated.uri)
                if fast_sha != associated.fast_sha:
                    bads.append(associated.uri)
                    if dry_run:  # Dry run
                        clean = 'n'
                    elif interactive:
                        clean = input("\tfast_sha for {} seems to have changed from {}"
                                      " to {}, do you wish to update?".format(
                                          associated.uri, associated.fast_sha, fast_sha))
                        if len(clean) > 0:
                            clean = clean[0]
                            clean = clean.lower()
                        else:
                            clean = 'y'
                    else:
                        clean = "y"
                    if clean == "y":
                        associated.fast_sha = fast_sha
        return bads

    @lock_strategies.lock_method
    def check_integrity(self):
        """Runs a sanity check on the dataset:
        1- Are associated files reachable?
        2- Did fast_shas change since file was associated
        """
        return self.cleanup_files(dry_run=True, clean_fastsha=True)

    @lock_strategies.lock_method
    def open(self, Id=None, loader=None, *args, **kargs):
        """open an object associated with a dataset

        :param Id: id of object to open, defaults to None which means first one.
        :type Id: str, optional
        :param loader: loader to use for this object, defaults to None
        :type loader: KoshLoader, optional
        :raises RuntimeError: object id not associated with dataset
        :return: object ready to be used
        """
        if Id is None:
            if len(self._associated_data_) > 0:
                Id = self._associated_data_[0]
            else:
                for Id in self._associated_data_:
                    return self.__store__.open(Id, loader)
        elif Id not in self._associated_data_:
            raise RuntimeError(
                "object {Id} is not associated with this dataset".format(
                    Id=Id))
        return self.__store__.open(Id, loader, *args, **kargs)

    @lock_strategies.lock_method
    def list_features(self, Id=None, loader=None,
                      use_cache=True, verbose=False, *args, **kargs):
        """list_features list features available if multiple associated data lead to duplicate feature name
        then the associated_data uri gets appended to feature name

        :param Id: id of associated object to get list of features from, defaults to None which means all
        :type Id: str, optional
        :param loader: loader to use to search for feature, will return ONLY features that the loader knows about
        :type loader: kosh.loaders.KoshLoader
        :param use_cache: If features is found on cache use it (default: True)
        :type use_cache: bool
        :param verbose: Verbose mode will show which file is being opened and errors on it
        :type verbose: bool
        :raises RuntimeError: object id not associated with dataset
        :return: list of features available
        :rtype: list
        """
        # Ok no need to sync any of this we will not touch the code
        saved_sync = self.__store__.is_synchronous()
        if saved_sync:
            # we will not update any rec in here, turning off sync
            # it makes things much faster
            backup = self.__store__.__sync__dict__
            self.__store__.__sync__dict__ = {}
            self.__store__.synchronous()
        features = []
        loaders = []
        associated_data = self._associated_data_
        if Id is None:
            for associated in associated_data:
                if verbose:
                    asso = self.__store__._load(associated)
                    print("Finding features for {}".format(asso))
                if loader is None:
                    ld, _ = self.__store__._find_loader(associated,
                                                        requestorId=self.id, verbose=verbose, use_cache=use_cache)
                else:
                    if (associated, self.id) not in self.__store__._cached_loaders:
                        self.__store__._cached_loaders[associated, self.id] = loader(
                            self.__store__._load(associated), requestorId=self.id), None
                        # self.__store__.update_cached_loaders()
                    ld, _ = self.__store__._cached_loaders[associated, self.id]
                loaders.append(ld)
                try:
                    features += ld._list_features(*
                                                  args, use_cache=use_cache, **kargs)
                except Exception as err:  # Ok the loader couldn't get the feature list
                    if verbose:
                        print("\tCould not obtain features from loader {}\n\t\tError:{}".format(loader, err))
            if len(features) != len(set(features)):
                # duplicate features we need to redo
                # Adding uri to feature name
                ided_features = []
                for index, associated in enumerate(associated_data):
                    ld = loaders[index]
                    if ld is None:
                        continue
                    asso = self.__store__._load(associated)
                    these_features = ld._list_features(
                        *args, use_cache=use_cache, **kargs)
                    for feature in these_features:
                        if features.count(feature) > 1:  # duplicate
                            ided_features.append(
                                "{feature}_@_{uri}".format(feature=feature, uri=asso.uri))
                        else:  # not duplicate name
                            ided_features.append(feature)
                features = ided_features
        elif Id not in self._associated_data_:
            raise RuntimeError(
                "object {Id} is not associated with this dataset".format(
                    Id=Id))
        else:
            if loader is not None:
                ld = loader
            else:
                ld, _ = self.__store__._find_loader(Id, requestorId=self.id, verbose=verbose)
            features = ld._list_features(*args, use_cache=use_cache, verbose=verbose, **kargs)
        features_id = self.__features__.get(Id, {})
        features_id[loader] = features
        self.__features__[Id] = features_id
        if saved_sync:
            # we need to restore sync mode
            self.__store__.__sync__dict__ = backup
            self.__store__.synchronous()
        return features

    @lock_strategies.lock_method
    def get_execution_graph(self, feature=None, Id=None,
                            loader=None, transformers=[], use_cache=True, *args, **kargs):
        """get data for a specific feature
        :param feature: feature (variable) to read, defaults to None
        :type feature: str, optional if loader does not require this
        :param Id: object to read in, defaults to None
        :type Id: str, optional
        :param loader: loader to use to get data,
                       defaults to None means pick for me
        :type loader: kosh.loaders.KoshLoader
        :param transformers: A list of transformers to use after the data is loaded
        :type transformers: kosh.transformer.KoshTranformer
        :param use_cache: use cache to find features
        :type use_cache: bool
        :returns: [description]
        :rtype: [type]
        """
        if feature is None:
            out = []
            for feat in self.list_features():
                out.append(self.get_execution_graph(Id=None,
                                                    feature=feat,
                                                    format=format,
                                                    loader=loader,
                                                    transformers=transformers,
                                                    *args, **kargs))
            return out
        # Need to make sure transformers are a list
        if not isinstance(transformers, Iterable):
            transformers = [transformers, ]
        # we need to figure which associated data has the feature
        if not isinstance(feature, list):
            features = [feature, ]
        else:
            features = feature
        possibles = {}
        inter = None
        union = set()

        alias_feature = self.alias_feature
        alias_feature_flattened = [[]] * len(alias_feature)
        for i, (key, val) in enumerate(alias_feature.items()):
            alias_feature_flattened[i] = [key]
            if isinstance(val, list):
                alias_feature_flattened[i].extend(val)
            elif isinstance(val, str):
                alias_feature_flattened[i].extend([val])

        def find_features(self, a, use_cache=True):

            a_original = a

            if "__uri__" in a:
                # Ok this is a pure sina file with mime_type
                a, _ = a.split("__uri__")
            a_obj = self.__store__._load(a)
            if loader is None:
                ld, _ = self.__store__._find_loader(a_original, requestorId=self.id, use_cache=use_cache)
                if ld is None:  # unknown mimetype probably
                    return _, None, _, _
            else:
                if a_obj.mime_type in loader.types:
                    ld = loader(a_obj, requestorId=self.id)
                else:
                    return _, None, _, _

            # Dataset with curve have themselves as uri
            obj_uri = getattr(ld.obj, "uri", "self")
            ld_features = ld._list_features(use_cache=use_cache)

            return a_original, ld, obj_uri, ld_features

        for index, feature_ in enumerate(features):
            possible_ids = []
            if Id is None:
                for a in self._associated_data_:

                    a_original, ld, obj_uri, ld_features = find_features(self, a, use_cache=use_cache)
                    if ld is None:  # No features
                        continue

                    if isinstance(ld, kosh.loaders.core.KoshSinaLoader):
                        # ok we have to be careful list_features can be returned two ways
                        if isinstance(ld_features[0], (list, tuple)) and isinstance(feature_, six.string_types):
                            # we need to convert the feature to str
                            possibilities = kosh.utils.find_curveset_and_curve_name(feature_, self.get_record())
                            if len(possibilities) > 1:
                                raise ValueError("cannot uniquely pinpoint {}, could be one of {}".format(
                                    feature_, possibilities))
                            feature_ = possibilities[0]
                            features[index] = feature_
                        elif isinstance(ld_features[0], six.string_types) and isinstance(feature_, (list, tuple)):
                            feature_ = "/".join(feature_)
                            features[index] = feature_

                    if ("_@_" not in feature_ and (feature_ in ld_features or list(feature_) in ld_features)) or\
                            feature_ is None or\
                            (feature_[:-len(obj_uri) - 3] in ld_features and
                             feature_[-len(obj_uri):] == obj_uri):
                        possible_ids.append(a_original)
                if possible_ids == []:  # All failed but could be something about the feature
                    found = False
                    if alias_feature_flattened:
                        feature_original = feature_
                        af = []
                        for af_list in alias_feature_flattened:
                            if feature_.split("/")[-1] in af_list:
                                af = af_list
                                if feature_ in af_list:
                                    af.remove(feature_)
                                break

                        poss = []
                        for a in self._associated_data_:

                            a_original, ld, obj_uri, ld_features = find_features(self, a, use_cache=use_cache)
                            if ld is None:  # No features
                                continue

                            for ld in ld_features:

                                if ld.split('/')[-1] in af or ld in af:

                                    found = True
                                    feature_ = ld
                                    features[index] = ld
                                    possible_ids.append(a_original)
                                    poss.append(ld)

                        if len(poss) > 1:
                            raise ValueError("Cannot uniquely pinpoint {}, could be one of {}"
                                             .format(feature_original, poss))
                    if not found:
                        raise ValueError("Cannot find feature {} in dataset"
                                         .format(feature_))
            elif Id == self.id:
                # Ok asking for data not associated externally
                # Likely curve
                ld, _ = self.__store__._find_loader(Id, requestorId=self.id, use_cache=use_cache)
                if feature_ in ld._list_features(use_cache=use_cache):
                    possible_ids = [Id, ]
                else:  # ok not a curve maybe a file?
                    rec = self.get_record()
                    for uri in rec["files"]:
                        if "mimetype" in rec["files"][uri]:
                            full_id = "{}__uri__{}".format(Id, uri)
                            ld, _ = self.__store__._find_loader(full_id, requestorId=self.id, use_cache=use_cache)
                            if ld is not None and feature_ in ld._list_features(use_cache=use_cache):
                                possible_ids = [full_id, ]
            elif Id not in self._associated_data_:
                raise RuntimeError(
                    "object {Id} is not associated with this dataset".format(
                        Id=Id))
            else:
                possible_ids = [Id, ]
            if inter is None:
                inter = set(possible_ids)
            else:
                inter = inter.intersection(set(possible_ids))
            union = union.union(set(possible_ids))
            possibles[feature_] = possible_ids

        if len(inter) != 0:
            union = inter

        ids = {}
        # Now let's go through each possible uri
        # and group features in them
        for id_ in union:
            matching_features = []
            for feature_ in features:
                if feature_ in possibles and id_ in possibles[feature_]:
                    matching_features.append(feature_)
                    del possibles[feature_]
            if len(matching_features) > 0:
                ids[id_] = matching_features

        out = []
        for id_ in ids:
            features = ids[id_]
            for Id in possible_ids:
                tmp = None
                try:
                    if loader is None:
                        ld, _ = self.__store__._find_loader(Id, requestorId=self.id, use_cache=use_cache)
                        mime_type = ld._mime_type
                    else:
                        if (Id, self.id) not in self.__store__._cached_loaders:
                            a_obj = self.__store__._load(Id)
                            self.__store__._cached_loaders[Id, self.id] = loader(a_obj, requestorId=self.id)
                            # self.__store__.update_cached_loaders()
                            mime_type = a_obj.mime_type
                        ld = self.__store__._cached_loaders[Id, self.id]
                    # Essentially make a copy
                    # Because we want to attach the feature to it
                    # But let's not lose the cached list_features
                    ld_uri = getattr(ld, "uri", None)
                    ld = ld.__class__(
                        ld.obj, mime_type=ld._mime_type, uri=ld_uri, requestorId=self.id)
                    # Ensures there is a possible path to format
                    get_graph(mime_type, ld, transformers)
                    final_features = []
                    obj_uri = getattr(ld.obj, "uri", "")
                    for feature_ in features:
                        if (feature_[:-len(obj_uri) - 3] in ld._list_features(use_cache=use_cache)
                                and feature_[-len(obj_uri):] == obj_uri):
                            final_features.append(
                                feature_[:-len(obj_uri) - 3])
                        else:
                            final_features.append(feature_)
                    if len(final_features) == 1:
                        final_features = final_features[0]
                    tmp = ld.get_execution_graph(final_features,
                                                 transformers=transformers
                                                 )
                    ld.feature = final_features
                    ExecGraph = kosh.exec_graphs.KoshExecutionGraph(tmp)
                except Exception:
                    import traceback
                    traceback.print_exc()
                    ExecGraph = kosh.exec_graphs.KoshExecutionGraph(tmp)
                out.append(ExecGraph)

        if len(out) == 1:
            return out[0]
        else:
            return out

    @lock_strategies.lock_method
    def get(self, feature=None, format=None, Id=None, loader=None,
            group=False, transformers=[], *args, **kargs):
        """get data for a specific feature
        :param feature: feature (variable) to read, defaults to None
        :type feature: str, optional if loader does not require this
        :param format: desired format after extraction
        :type format: str
        :param Id: object to read in, defaults to None
        :type Id: str, optional
        :param loader: loader to use to get data,
                       defaults to None means pick for me
        :type loader: kosh.loaders.KoshLoader
        :param group: group multiple features in one get call, assumes loader can handle this
        :type group: bool
        :param transformers: A list of transformers to use after the data is loaded
        :type transformers: kosh.transformer.KoshTranformer
        :raises RuntimeException: could not get feature
        :raises RuntimeError: object id not associated with dataset
        :returns: [description]
        :rtype: [type]
        """
        G = self.get_execution_graph(
            feature=feature,
            Id=Id,
            loader=loader,
            transformers=transformers,
            *args,
            **kargs)
        if isinstance(G, list):
            return [g.traverse(format=format, *args, **kargs) for g in G]
        else:
            return G.traverse(format=format, *args, **kargs)

    @lock_strategies.lock_method
    def __getitem__(self, feature):
        """Shortcut to access a feature or list of
        :param feature: feature(s) to access in dataset
        :type feature: str or list of str
        :returns: (list of) access point to feature requested
        :rtype: (list of) kosh.execution_graph.KoshIoGraph
        """
        return self.get_execution_graph(feature)

    @lock_strategies.lock_method
    def __dir__(self):
        """__dir__ list functions and attributes associated with dataset
        :return: functions, methods, attribute associated with this dataset
        :rtype: list
        """
        current = set(super(KoshDataset, self).__dir__())
        try:
            atts = set(self.listattributes() + self.__protected__)
        except Exception:
            atts = set()
        return list(current.union(atts))

    @lock_strategies.lock_method
    def reassociate(self, target, source=None, absolute_path=True):
        """This function allows to re-associate data whose uri might have changed

        The source can be the original uri or sha and target is the new uri to use.
        :param target: New uri
        :type target: str
        :param source: uri or sha (long or short of reassociate)
                       to reassociate with target, if None then the short uri from target will be used
        :type source: str or None
        :param absolute_path: if file exists should we store its absolute_path
        :type absolute_path: bool
        :return: None
        :rtype: None
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        self.__store__.reassociate(target, source=source, absolute_path=absolute_path)

    @lock_strategies.lock_method
    def validate(self):
        """If dataset has a schema then make sure all attributes pass the schema"""
        if self.schema is not None:
            self.schema.validate(self)

    @lock_strategies.lock_method
    def searchable_source_attributes(self):
        """Returns all the attributes of associated sources
        :return: List of all attributes you can use to search sources in the dataset
        :rtype: set
        """
        searchable = set()
        for source in self.find():
            searchable = searchable.union(source.listattributes())
        return searchable

    @lock_strategies.lock_method
    def describe_feature(self, feature, Id=None, **kargs):
        """describe a feature

        :param feature: feature (variable) to read, defaults to None
        :type feature: str, optional if loader does not require this
        :param Id: id of associated object to get list of features from, defaults to None which means all
        :type Id: str, optional
        :param kargs: keywords to pass to list_features (optional)
        :type kargs: keyword=value
        :raises RuntimeError: object id not associated with dataset
        :return: dictionary describing the feature
        :rtype: dict
        """
        loader = None
        if Id is None:
            for a in self._associated_data_:
                ld, _ = self.__store__._find_loader(a, requestorId=self.id)
                if feature in ld._list_features(**kargs) or \
                        (feature[:-len(ld.obj.uri) - 3] in ld._list_features()
                         and feature[-len(ld.obj.uri):] == ld.obj.uri):
                    loader = ld
                    break
        elif Id not in self._associated_data_:
            raise RuntimeError(
                "object {Id} is not associated with this dataset".format(
                    Id=Id))
        else:
            loader, _ = self.__store__._find_loader(Id, requestorId=self.id)
        return loader.describe_feature(feature)

    @lock_strategies.lock_method
    def dissociate(self, uri, absolute_path=True):
        """dissociates a uri/mime_type with this dataset

        :param uri: uri to access file
        :type uri: str
        :param absolute_path: if file exists should we store its absolute_path
        :type absolute_path: bool
        :return: None
        :rtype: None
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        if absolute_path and os.path.exists(uri):
            uri = os.path.abspath(uri)
        rec = self.get_record()
        if uri not in rec["files"]:
            # Not associated with this uri anyway
            return
        kosh_id = str(rec["files"][uri]["kosh_id"])
        del rec["files"][uri]
        now = time.time()
        rec["user_defined"]['kosh_information']["{uri}___associated_last_modified".format(
            uri=uri)] = now
        if self.__store__.__sync__:
            self._update_record(rec)
        # Get all object that have been associated with this uri
        rec = self.__store__.get_record(kosh_id)
        associated_ids = rec.data.get("associated", {"value": []})["value"]
        associated_ids.remove(self.id)
        rec.data["associated"]["value"] = associated_ids
        if self.__store__.__sync__:
            self._update_record(rec)
        else:
            self._update_record(rec, self.__store__._added_unsync_mem_store)
        if len(associated_ids) == 0:  # ok no other object is associated
            self.__store__.delete(kosh_id)
            if (kosh_id, self.id) in self.__store__._cached_loaders:
                del self.__store__._cached_loaders[kosh_id, self.id]
            if (kosh_id, None) in self.__store__._cached_loaders:
                del self.__store__._cached_loaders[kosh_id, None]
            # We also need to clean up all features associated with the id
            cached_features = self.__store__._cached_features_
            yank = []
            for id in cached_features:
                if id.endswith(uri):
                    yank.append(id)
            for id in yank:
                del cached_features[id]
            self.__store__._cached_features_ = cached_features

        # Since we changed the associated, we need to cleanup
        # the features cache
        self.__features__[None] = {}
        self.__features__[kosh_id] = {}

    @lock_strategies.lock_method
    def associate(self, uri, mime_type, metadata={},
                  id_only=None, long_sha=False, absolute_path=True,
                  loader_kwargs=None, preload_features=False):
        """associates a uri/mime_type with this dataset

        :param uri: uri(s) to access content
        :type uri: str or list of str
        :param mime_type: mime type associated with this file
        :type mime_type: str or list of str
        :param metadata: metadata to associate with file, defaults to {}
        :type metadata: dict, optional
        :param id_only: do not return kosh file object (None and default), just its id (True)
                        or the KoshSinaFile (False)
        :type id_only: bool or None
        :param long_sha: Do we compute the long sha on this or not?
        :type long_sha: bool
        :param absolute_path: if file exists should we store its absolute_path
        :type absolute_path: bool
        :param loader_kwargs: Extra arguments to pass to more advanced loader
        :type loader_kwargs: dict, optional
        :param preload_features: runs list_features on associated uri to save time in future reads (default: False)
        :type preload_features: bool
        :return: A (list) Kosh Sina File(s)
        :rtype: list of KoshSinaFile or KoshSinaFile
        """
        from sina.model import Record
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        rec = self.get_record()
        # Need to remember we touched associated files
        now = time.time()

        if loader_kwargs is not None:
            pickled = kosh_pickler.dumps(loader_kwargs)
            metadata['loader_kwargs'] = pickled

        if isinstance(uri, six.string_types):
            uris = [uri, ]
            metadatas = [metadata, ]
            mime_types = [mime_type, ]
            single_element = True
        else:
            uris = uri
            if isinstance(metadata, dict):
                metadatas = [metadata, ] * len(uris)
            else:
                metadatas = metadata
            if isinstance(mime_type, six.string_types):
                mime_types = [mime_type, ] * len(uris)
            else:
                mime_types = mime_type
            single_element = False

        new_recs = []
        updated_recs = []
        kosh_file_ids = []

        for i, uri in enumerate(uris):
            try:
                meta = metadatas[i].copy()
                if os.path.exists(uri):
                    if long_sha:
                        meta["long_sha"] = compute_long_sha(uri)
                    if absolute_path:
                        uri = os.path.abspath(uri)
                    if not os.path.isdir(uri) and "fast_sha" not in meta:
                        meta["fast_sha"] = compute_fast_sha(uri)
                rec["user_defined"]['kosh_information']["{uri}___associated_last_modified".format(
                    uri=uri)] = now
                # We need to check if the uri was already associated somewhere
                tmp_uris = list(self.__store__.find(
                    types=[self.__store__._sources_type, ], uri=uri, ids_only=True))

                if len(tmp_uris) == 0:
                    Id = uuid.uuid4().hex
                    rec_obj = Record(id=Id, type=self.__store__._sources_type, user_defined={'kosh_information': {}})
                    new_recs.append(rec_obj)
                else:
                    rec_obj = self.__store__.get_record(tmp_uris[0])
                    Id = rec_obj.id
                    existing_mime = rec_obj["data"]["mime_type"]["value"]
                    mime_type = mime_types[i]
                    if existing_mime != mime_types[i]:
                        rec["files"][uri]["mime_type"] = existing_mime
                        raise TypeError("source {} is already associated with another dataset with mimetype"
                                        " '{}' you specified mime_type '{}'".format(uri, existing_mime, mime_types[i]))
                    updated_recs.append(rec_obj)
                rec.add_file(uri, mime_types[i])
                rec["files"][uri]["kosh_id"] = Id
                meta["uri"] = uri
                meta["mime_type"] = mime_types[i]
                associated = rec_obj["data"].get(
                    "associated", {'value': []})["value"]
                associated.append(self.id)
                meta["associated"] = associated
                for key in meta:
                    if key not in rec_obj.data:
                        rec_obj.add_data(key, meta[key])
                    else:
                        rec_obj["data"][key]["value"] = meta[key]
                    last_modif_att = "{name}_last_modified".format(name=key)
                    rec_obj["user_defined"]['kosh_information'][last_modif_att] = time.time()
                if not self.__store__.__sync__:
                    rec_obj["user_defined"]['kosh_information']["last_update_from_db"] = time.time()
                    self.__store__.__sync__dict__[Id] = rec_obj
            except TypeError as err:
                raise err
            except Exception:
                # file already in there
                # Let's get the matching id
                if rec_obj["data"]["mime_type"]["value"] != mime_types[i]:
                    raise TypeError("file {} is already associated with this dataset with mimetype"
                                    " '{}' you specified mime_type '{}'".format(uri, existing_mime, mime_type))
                else:
                    Id = rec["files"][uri]["kosh_id"]
                    if len(metadatas[i]) != 0:
                        warnings.warn(
                            "uri {} was already associated, metadata will "
                            "stay unchanged\nEdit object (id={}) directly to update attributes.".format(uri, Id))
            kosh_file_ids.append(Id)

        if self.__store__.__sync__:
            self.__store__.lock()
            self.__store__.__record_handler__.insert(new_recs)
            self.__store__.unlock()
            self._update_record(rec)
            for rec_up in updated_recs:
                self._update_record(rec_up)
        else:
            self._update_record(rec, self.__store__._added_unsync_mem_store)
            for rec_up in updated_recs:
                self._update_record(rec_up, self.__store__._added_unsync_mem_store)

        # Since we changed the associated, we need to cleanup
        # the features cache
        self.__features__[None] = {}
        # Let's rerun the list_features now so it's cached in the store and for the dataset.
        if preload_features:
            self.list_features()

        if id_only is None:
            return

        if id_only:
            if single_element:
                return kosh_file_ids[0]
            else:
                return kosh_file_ids

        kosh_files = []
        for Id in kosh_file_ids:
            self.__features__[Id] = {}
            kosh_file = KoshSinaObject(Id=Id,
                                       kosh_type=self.__store__._sources_type,
                                       store=self.__store__,
                                       metadata=metadata,
                                       record_handler=self.__record_handler__)
            kosh_files.append(kosh_file)

        if single_element:
            return kosh_files[0]
        else:
            return kosh_files

    @lock_strategies.lock_method
    def search(self, *atts, **keys):
        """
        Deprecated use find
        """
        warnings.warn("The 'search' function is deprecated and now called `find`.\n"
                      "Please update your code to use `find` as `search` might disappear in the future",
                      DeprecationWarning)
        return self.find(*atts, **keys)

    @lock_strategies.lock_method
    def find(self, *atts, **keys):
        """find associated data matching some metadata
        arguments are the metadata name we are looking for e.g
        find("attr1", "attr2")
        you can further restrict by specifying exact value for a metadata
        via key=value
        you can return ids only by using: ids_only=True
        range can be specified via: sina.utils.DataRange(min, max)

        "file_uri" is a special key that will return the kosh object associated
        with this dataset for the given uri.  e.g store.find(file_uri=uri)

        :return: list of matching objects associated with dataset
        :rtype: list
        """
        from sina.utils import exists

        if self._associated_data_ is None:
            return
        sina_kargs = {}
        ids_only = keys.pop("ids_only", False)
        load_type = keys.pop("load_type", "source")
        # We are only interested in ids from Sina
        sina_kargs["ids_only"] = True

        inter_recs = self._associated_data_
        tag = "{}__uri__".format(self.id)
        tag_len = len(tag)
        virtuals = [x[tag_len:] for x in inter_recs if x.startswith(tag)]
        # Bug in sina 1.10.0 forces us to remove the virtual from the pool
        for v_id in virtuals:
            inter_recs.remove("{}{}".format(tag, v_id))
        if "file" in sina_kargs and "file_uri" in sina_kargs:
            raise ValueError(
                "The `file` keyword is being deprecated for `file_uri` you cannot use both")
        if "file" in sina_kargs:
            warnings.warn(
                "The `file` keyword has been renamed `file_uri` and may not be available in future versions",
                DeprecationWarning)
            file_uri = sina_kargs.pop("file")
            sina_kargs["file_uri"] = file_uri

        # The data dict for sina
        keys.pop("id_pool", None)
        sina_kargs["query_order"] = keys.pop(
            "query_order", ("data", "file_uri", "types"))
        sina_data = keys.pop("data", {})
        for att in atts:
            sina_data[att] = exists()
        sina_data.update(keys)
        sina_kargs["data"] = sina_data

        match = set(
            self.__record_handler__.find(
                id_pool=inter_recs,
                **sina_kargs))
        # instantly restrict to associated data
        if not self.__store__.__sync__:
            match_mem = set(self.__store__._added_unsync_mem_store.records.find(
                id_pool=inter_recs, **sina_kargs))
            match = match.union(match_mem)
        # ok now we need to search the data on the virtual datasets
        rec = self.get_record()
        for uri in virtuals:
            file_section = rec["files"][uri]
            tags = file_section.get("tags", [])
            # Now let's search the tags....
            match_it = True
            for key in sina_kargs["data"]:
                if key == "mime_type":
                    if file_section.get("mimetype", file_section.get(
                            "mime_type", None)) != sina_kargs["data"][key]:
                        match_it = False
                        break
                elif key not in tags or sina_kargs["data"][key] != exists():
                    match_it = False
                    break
            if match_it:
                # we can't have a set anymore
                match.add(self.id)

        for rec_id in match:
            # We need to cleanup for "virtual association".
            # e.g comes directly from a sina rec with 'file'/'mimetype' in it.
            rec_id = rec_id.split("__uri__")[0]
            if load_type == 'source':
                yield rec_id if ids_only else self.__store__._load(rec_id)
            elif load_type == 'dictionary':
                yield rec_id if ids_only else self.__record_handler__.get(rec_id).__dict__['raw']

    @lock_strategies.lock_method
    def export(self, file=None, sina_record=False):
        """Exports this dataset
        :param file: export dataset to a file
        :type file: None or str
        :param sina_record: export the dataset as a Sina record
        :type sina_record: bool
        :return: dataset and its associated data
        :rtype: dict"""
        rec = self.get_record()
        relationships = self.get_sina_store().relationships.find(self.id)
        # cleanup the record
        rec_json = cleanup_sina_record_from_kosh_sync(rec)
        jsns = [rec_json, ]
        # ok now same for associated data
        for associated_id in self._associated_data_:
            if associated_id.startswith(f"{self.id}__uri__"):
                # ok self referencing no need to add
                continue
            rec = self.__store__._load(associated_id).get_record()
            rec_json = cleanup_sina_record_from_kosh_sync(rec)
            jsns.append(rec_json)

        # returns a dict that should be ingestible by sina
        output_dict = {
            "minimum_kosh_version": None,
            "kosh_version": kosh.version(comparable=True),
            "sources_type": self.__store__._sources_type,
            "records": jsns,
            "relationships": relationships
        }

        update_json_file_with_records_and_relationships(file, output_dict)
        return output_dict

    @lock_strategies.lock_method
    def get_associated_data(self, ids_only=False):
        """Generator of associated data
        :param ids_only: generator will return ids if True Kosh object otherwise
        :type ids_only: bool
        :returns: generator
        :rtype: str or Kosh objects
        """
        for id in self._associated_data_:
            if ids_only:
                yield id
            else:
                yield self.__store__._load(id)

    @lock_strategies.lock_method
    def is_member_of(self, ensemble):
        """Determines if this dataset is a member of the passed ensemble
        :param ensemble: ensemble we need to determine if this dataset is part of
        :type ensemble: str or KoshEnsemble

        :returns: True if member of the ensemble, False otherwise
        :rtype: bool"""
        if not isinstance(ensemble, (six.string_types, kosh.ensemble.KoshEnsemble)):
            raise TypeError("ensemble must be id or KoshEnsemble object")
        if isinstance(ensemble, kosh.ensemble.KoshEnsemble):
            ensemble = ensemble.id

        return ensemble in self.get_ensembles(ids_only=True)

    @lock_strategies.lock_method
    def get_ensembles(self, ids_only=False):
        """Returns the ensembles this dataset is part of
        :param ids_only: return ids or objects
        :type ids_only: bool
        """
        for rel in self.get_sina_store().relationships.find(
                self.id, self.__store__._ensemble_predicate, None):
            if ids_only:
                yield rel.object_id
            else:
                yield self.__store__.open(rel.object_id, requestorId=self.id)

    @lock_strategies.lock_method
    def leave_ensemble(self, ensemble):
        """Removes this dataset from an ensemble
        :param ensemble: The ensemble to leave
        :type ensemble: str or KoshEnsemble
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        from kosh.ensemble import KoshEnsemble
        if isinstance(ensemble, six.string_types):
            ensemble = self.__store__.open(ensemble)
        if not isinstance(ensemble, KoshEnsemble):
            raise ValueError(
                "cannot join `ensemble` since object `{}` does not map to an ensemble".format(ensemble))
        if self.id in ensemble.get_members(ids_only=True):
            ensemble.remove(self.id)
        else:
            warnings.warn(
                "{} is not part of ensemble {}. Ignoring request to leave it.".format(
                    self.id, ensemble.id))

    @lock_strategies.lock_method
    def join_ensemble(self, ensemble):
        """Adds this dataset to an ensemble
        :param ensemble: The ensemble to join
        :type ensemble: str or KoshEnsemble
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        from kosh.ensemble import KoshEnsemble
        if isinstance(ensemble, six.string_types):
            ensemble = self.__store__.open(ensemble, requestorId=self.id)
        if not isinstance(ensemble, KoshEnsemble):
            raise ValueError(
                "cannot join `ensemble` since object `{}` does not map to an ensemble".format(ensemble))
        ensemble.add(self)

    @lock_strategies.lock_method
    def clone(self, preserve_ensembles_memberships=False, id_only=False):
        """Clones the dataset, e.g makes an identical copy.

        :param preserve_ensembles_memberships: Add the new dataset to the ensembles this original dataset belongs to?
                                               True/1:  The cloned dataset will belong to the same ensembles
                                                        as the original dataset.
                                               False/0: The new dataset will not belong to any ensemble
                                                        but we will copy ensemble level attributes onto
                                                        the cloned dataset.
                                               -1:      The new dataset will not belong to any ensemble and we
                                                        will leave out all attributes that belong to ensembles.
        :type preserve_ensembles_membership: bool or int

        :param id_only: returns id rather than new dataset
        :type id_only: bool

        :returns: The cloned dataset or its id
        :rtype: KoshDataset or str
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        attributes = self.list_attributes(True)
        cloned_dataset = self.__store__.create(metadata=attributes)
        for associated in self.get_associated_data():
            rec = associated.get_record()
            if "uri" not in rec["data"]:
                continue
            cloned_dataset.associate(
                associated.uri,
                associated.mime_type,
                metadata=associated.list_attributes(True))

        if preserve_ensembles_memberships in [True, 1]:
            for ensemble in self.get_ensembles():
                ensemble.add(cloned_dataset)
        elif preserve_ensembles_memberships == -1:
            for attribute in self.list_attributes():
                if self.is_ensemble_attribute(attribute):
                    delattr(cloned_dataset, attribute)
        elif preserve_ensembles_memberships != 0:
            raise ValueError(
                "preserve_ensembles_memberships must be one of True/1, False/0 or -1")

        if id_only:
            return cloned_dataset.id
        else:
            return cloned_dataset

    @lock_strategies.lock_method
    def is_ensemble_attribute(self, attribute, ensembles=None, ensemble_id=False):
        """Determine if an attribute belongs to ensemble this dataset is part of
        :param attribute: The attribute to check
        :type attribute: str
        :param ensembles: ensembles to check against, defaults to all ensembles the dataset belongs to
        :type ensembles: None, ensemble_id or ensemble_obj or list of these
        :param ensemble_id: rather than True/False return the id of the ensemble this attribute comes from
        :type ensemble_id: bool
        :returns: True or ensemble of origin id if the attribute belongs to an ensemble False/"" otherwise
        """
        if ensembles is None:
            ensembles = self.get_ensembles()
        elif isinstance(ensembles, str):
            try:
                ensembles = self.__store__.open(ensembles, requestorId=self.id)
            except Exception:
                raise ValueError(
                    "could not find object with id {} in the store".format(ensembles))
            ensembles = [ensembles, ]
        elif not isinstance(ensembles, Iterable):
            ensembles = [ensembles, ]
        for ensemble in ensembles:
            if not isinstance(ensemble, kosh.ensemble.KoshEnsemble):
                raise ValueError(
                    "Object with id {} is not an ensemble {}".format(
                        ensemble.id, type(ensemble)))

            if attribute in ensemble.list_attributes(no_duplicate=True):
                if ensemble_id:
                    return ensemble.id
                else:
                    return True
        if ensemble_id:
            return ""
        else:
            return False

    @lock_strategies.lock_method
    def list_ensemble_tags(self, ensemble_id, dictionary=False, obscure=True):
        """list all ensemble tags of specific ensemble ids

        :param ensemble_id: Ensemble ID(s) of ensemble(s)
        :type ensemble_id: str, str
        :param dictionary: return a dictionary of value/pair rather than just tag names
        :type dictionary: bool
        :param obscure: Don't return backend attributes such as 'INHERIT_ATTRIBUTES'
        :type obscure: bool

        :return: list of ensemble tags for dataset of a specific ensemble
        :rtype: list
        """
        ens_tags = self.list_attributes(ensemble_id=ensemble_id, dictionary=dictionary, obscure=obscure)
        if dictionary:
            return {key: val for key, val in ens_tags.items() if "_ENSEMBLE_TAG_" in key}
        else:
            return [et for et in ens_tags if "_ENSEMBLE_TAG_" in et]

    @lock_strategies.lock_method
    def add_ensemble_tags(self, ensemble_id, ensemble_tags):
        """add ensemble tags to a specific ensemble

        :param ensemble_id: Ensemble ID of ensemble
        :type ensemble_id: str
        :param ensemble_tags: Ensemble tags and their values to add
        :type ensemble_tags: dict
        """
        for key, val in ensemble_tags.items():
            self.___setattr___(f"{ensemble_id}_ENSEMBLE_TAG_{key}", val, force=True)

    @lock_strategies.lock_method
    def delete_ensemble_tags(self, ensemble_id, ensemble_tags):
        """remove ensemble tags from a specific ensemble

        :param ensemble_id: Ensemble ID of ensemble
        :type ensemble_id: str
        :param ensemble_tags: Ensemble tags to remove
        :type ensemble_tags: list
        """
        if isinstance(ensemble_tags, str):
            ensemble_tags = [ensemble_tags]
        for et in ensemble_tags:
            attr = f"{ensemble_id}_ENSEMBLE_TAG_{et}"
            delattr(self, attr)

    @lock_strategies.lock_method
    def add_curve(self, curve, curve_set=None, curve_name=None, independent=None, units=None, tags=None):
        """Add a curve to a dataset
        :param curve: The curve data
        :type curve: array-like
        :param curve_set: Name of the curve_set.
                          If not passed will be set to curve_x where x is the highest number available
                          (if non exist then curve_set_0 will be created)
        :type curve_set: str
        :param curve_name: name of the curve.
                           If not set it will be set to: curve_x where x is the number of curves in this curve_set
        :type curve_name: str
        :param independent: Is it an independent variable. If not passed, it will be set to False
                            unless no independent curve exists, e.g first curve is set as independent
                            You can avoid this by explicitly passing False
        :type independent: bool
        :param units: Units of the curve (if any)
        :type units: str
        :param tags: Tags on the curve (if any)
        :type tags: dict

        :returns: the curve_set/curve_name path
        :rtype: str
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        # Before anything else lets remove the cached features
        cached_features = self.__store__._cached_features_
        yank = []
        for id in cached_features:
            if id.endswith(self.id):
                yank.append(id)
        for id in yank:
            del cached_features[id]
        # put back in store
        self.__store__._cached_features_ = cached_features

        # let's obtain the record
        rec = self.get_record()
        # Let's figure out the curve_set name
        if curve_set is None:
            i = 0
            while "curve_set_"+str(i) in rec.raw["curve_sets"]:
                i += 1
            if i > 0:
                i -= 1
            curve_set = "curve_set_"+str(i)

        # Let's create curveset if not present
        if curve_set not in rec.raw["curve_sets"]:
            cs = rec.add_curve_set(curve_set)
        else:
            cs = rec.get_curve_set(curve_set)
        # Ok now let's get a curve_name
        if curve_name is None:
            i = 0
            while "curve_"+str(i) in cs.dependent or "curve_"+str(i) in cs.independent:
                i += 1
            curve_name = "curve_"+str(i)

        if curve_name in cs.independent or curve_name in cs.dependent:
            raise ValueError("curve {} already exist in curve_set {}".format(curve_name, curve_set))
        # Finally the independent part
        if independent is None:
            if len(cs.independent) == 0:
                independent = True
            else:
                independent = False
        if independent:
            cs.add_independent(curve_name, curve, units, tags)
        else:
            cs.add_dependent(curve_name, curve, units, tags)

        if self.__store__.__sync__:
            self._update_record(rec)
        else:
            self._update_record(rec, self.__store__._added_unsync_mem_store)

        # Since we changed the curves, we need to cleanup
        # the features cache
        self.__features__[None] = {}

        return "{}/{}".format(curve_set, curve_name)

    @lock_strategies.lock_method
    def remove_curve_or_curve_set(self, curve, curve_set=None):
        """Removes a curve or curve_set from the dataset
        :param curve: name of the curve or curve_set to remove
        :type curve: str
        :param curve_set: curve_set the curve_name belongs to.
                          If not passed then assumes it is in the curve
                          name.
        :type curve_set: str
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        original_curve_set = curve_set
        rec = self.get_record()
        if curve_set is None:
            # User did not pass the curve_set
            # The curve_set might be prepended to the curve_name then.
            try:
                curve_set, curve = kosh.utils.find_curveset_and_curve_name(curve, rec)[0]
            except Exception:
                raise ValueError("You need to pass a curve set for curve {}".format(curve))
        if curve_set not in rec.raw["curve_sets"]:
            if original_curve_set is None:
                if curve in rec.raw["curve_sets"]:
                    del rec.raw["curve_sets"][curve]
                    if self.__store__.__sync__:
                        self._update_record(rec)
                    else:
                        self._update_record(rec, self.__store__._added_unsync_mem_store)
                    self.__features__[None] = {}
                    return
            else:
                raise ValueError("The curve set {} does not exists".format(curve_set))
        cs = rec.get_curve_set(curve_set)
        if curve in cs.independent:
            del rec.raw["curve_sets"][curve_set]["independent"][curve]
        elif curve in cs.dependent:
            del rec.raw["curve_sets"][curve_set]["dependent"][curve]
        elif curve is None:
            # We want to delete the whole curveset
            del rec.raw["curve_sets"][curve_set]
        else:
            raise ValueError("Could not find {} in curve_set {}".format(curve, curve_set))
        # In case we removed everything
        if len(cs.dependent) == 0 and len(cs.independent) == 0:
            del rec.raw["curve_sets"][curve_set]

        if self.__store__.__sync__:
            self._update_record(rec)
        else:
            self._update_record(rec, self.__store__._added_unsync_mem_store)

        # Since we changed the curves, we need to cleanup
        # the features cache
        self.__features__[None] = {}
        return

    @lock_strategies.lock_method
    def list_attributes(self, dictionary=False, ensemble_id=None, obscure=True):
        """listattributes list all non protected attributes

        :param dictionary: return a dictionary of value/pair rather than just attributes names
        :type dictionary: bool
        :param ensemble_id: Provide ensemble ID(s) to return ensemble tags
        :type ensemble_id: str, lst
        :param obscure: Don't return backend attributes such as 'INHERIT_ATTRIBUTES'
        :type obscure: bool

        :return: list of attributes set on object
        :rtype: list
        """

        attributes = super(KoshDataset, self).list_attributes(dictionary=dictionary, ensemble_id=ensemble_id)
        if obscure:
            if dictionary:
                return {key: val for key, val in attributes.items() if '_ENSEMBLE_TAG_INHERIT_ATTRIBUTES' not in key}
            else:
                return [et for et in attributes if '_ENSEMBLE_TAG_INHERIT_ATTRIBUTES' not in et]
        return attributes

    @lock_strategies.lock_method
    def to_dataframe(self, data_columns=[], *atts, **keys):
        """Return the find object as a Pandas DataFrame.

        Pass in the same arguments and keyword arguments as the find method.

        find associated data matching some metadata
        arguments are the metadata name we are looking for e.g
        find("attr1", "attr2")
        you can further restrict by specifying exact value for a metadata
        via key=value
        you can return ids only by using: ids_only=True
        range can be specified via: sina.utils.DataRange(min, max)

        "file_uri" is a special key that will return the kosh object associated
        with this dataset for the given uri.  e.g store.find(file_uri=uri)

        :param data_columns: Columns to extract. By default this will include ['id', 'mime_type', 'uri', 'associated'].
                            If nothing is passed, will return all data.
        :type data_columns: Union(str, list), optional
        :return: Pandas DataFrame
        :rtype: Pandas DataFrame
        """
        import pandas as pd
        if isinstance(data_columns, str):
            data_columns = [data_columns]

        keys['load_type'] = 'dictionary'
        keys['ids_only'] = False
        sources = list(self.find(*atts, **keys))

        attr_dict = {}
        total_sources = len(sources)

        # Always have these by default
        defaults = ['id', 'mime_type', 'uri', 'associated']

        # Acquire all data if `data_columns` was not passed
        if not data_columns:
            unique_keys = []
            for i, source in enumerate(sources):
                unique_keys.extend(list(source['data'].keys()))

            data_columns = sorted(set(unique_keys))

        data_columns = defaults + data_columns  # Want defaults in front
        attr_dict = {d: [pd.NA] * total_sources for d in data_columns}

        for i, source in enumerate(sources):
            for column in data_columns:
                if column == "id":
                    attr_dict[column][i] = source['id']
                else:
                    attr_dict[column][i] = source['data'].get(column, {}).get('value', pd.NA)

        df = pd.DataFrame(attr_dict)
        return df



from ..exec_graphs import KoshExecutionGraph
from abc import abstractmethod
import hashlib
from kosh import kosh_cache_dir
import os
import functools


class KoshOperator(KoshExecutionGraph):
    # TODO something about operator accepting multiple types in (list?)
    # For now assuming it is type of first input received
    types = {}

    def __init__(self,
                 *args, **kargs):
        """init function
        output signature is generated from the input args (w/o the cache_dir)
        :param cache_dir: directory to save cached files must be passed as key/value
        :type cache_dir: str
        :param cache: do we use cache? 0: no, 1:yes, 2:yes but clobber if exists must be passed as key/value
        :type cache: int
        :param verbose: Turn on verbosity, by default this will turn on printing a message
                        when results are loaded from cache. Message is sent as lone argument
                        to `self._print` function.
                        value is stored in self._verbose
        :type verbose: bool
        """
        self.cache_dir = kargs.pop("cache_dir", kosh_cache_dir)
        self.signature = hashlib.sha256(repr(self.__class__).encode()).hexdigest()
        self.signature = self.update_signature(*args, **kargs)
        self.use_cache = kargs.pop("use_cache", False)
        self._verbose = kargs.pop("verbose", False)
        cache = kargs.pop("cache", False)
        if cache:
            try:
                os.makedirs(self.cache_dir)
            except Exception:
                pass
        self.cache = cache
        super(KoshOperator, self).__init__(*args, **kargs)

    def operate_(self, *inputs, **kargs):
        """Given input(s) from previous loader, transformer or operators and desired format
        computes the unique signature and tries to extract from cache, calls operator's
        `operate` function if no cache available.
        :param *inputs: set of input passed from loader or previous transformer
        :type *inputs: object
        :param format: desired output format, must be passed as keyword
                       will be attached to object for retrieval in the function
        :type format: str
        :return: The result from operate function on inputs
        :rtype: object
        """

        format = kargs["format"]
        if self.cache:
            signature = kargs.get("signature", None)

            if signature is None:
                use_signature = self.update_signature(
                    inputs, format)
            else:
                use_signature = signature

            cache_file = os.path.join(self.cache_dir, use_signature)
            if self.cache == 2 and os.path.exists(cache_file):
                # User wants to clobber cache
                os.remove(cache_file)

            try:
                result = self.load(use_signature)
                if self._verbose:
                    self._print("Loaded results from cache file {} using signature: {}".format(
                        cache_file, use_signature))
            except Exception:
                if signature is None:
                    signature = self.update_signature(
                        inputs, format)
                result = self.operate(*inputs, format=format)
                if self.cache > 0:  # Ok user wants to cache results
                    if not os.path.exists(self.cache_dir):
                        os.makedirs(self.cache_dir)
                    self.save(signature, result)
        else:
            result = self.operate(*inputs, format=format)

        return result

    def _print(self, message):
        print(message)

    @abstractmethod
    def operate(self, *inputs, **kargs):
        """The operating function on the inputs
        :param inputs: result returned by loader or previous transformer
        :type inputs: tuple of features/execution graphs
        """
        raise NotImplementedError("the operate function is not implemented")


def typed_operator_with_kwargs(types_dict=None):
    if types_dict is None:
        types_dict = {"numpy": ["numpy", ]}

    def make_operator(*inputs, **kargs):
        def actual_operator_decorator(func):
            class CustomOperator(KoshOperator):
                types = types_dict

                @functools.wraps(func)
                def operate(self, *operate_inputs, **operate_kargs):
                    return func(*operate_inputs, **operate_kargs)

                def __getitem_propagate__(self, key, input_index):
                    return key
            return CustomOperator
        return actual_operator_decorator(*inputs, **kargs)
    return make_operator


def typed_operator(types_dict=None):
    if types_dict is None:
        types_dict = {"numpy": ["numpy", ]}

    def make_operator(*inputs, **kargs):
        def actual_operator_decorator(func):
            class CustomOperator(KoshOperator):
                types = types_dict

                @functools.wraps(func)
                def operate(self, *operate_inputs, **operate_kargs):
                    return func(*operate_inputs)

                def __getitem_propagate__(self, key, input_index):
                    return key
            return CustomOperator
        return actual_operator_decorator(*inputs, **kargs)
    return make_operator


numpy_operator = typed_operator()



import uuid
import warnings
import time
import reprlib
from .schema import KoshSchema
from .utils import KoshPickler
from . import lock_strategies


kosh_pickler = KoshPickler()


def __getattr__(name):
    if name == "sina_version":
        from sina import get_version
        return float(".".join(get_version().split(".")[:2]))


class KoshSinaObject(object):
    """KoshSinaObject Base class for sina objects
    """
    @lock_strategies.lock_method
    def get_record(self):
        return self.__store__.get_record(self.id)

    def __init__(self, Id, store, kosh_type,
                 record_handler, protected=[], metadata={}, schema=None,
                 record=None):
        from sina.model import Record
        """__init__ sina object base class

        :param Id: id to use for unique identification, if None is passed set for you via uui4()
        :type Id: str
        :param store: Kosh store associated
        :type store: KoshSinaStore
        :param kosh_type: type of Kosh object (dataset, file, project, ...)
        :type kosh_type: str
        :param record_handler: sina record handler object
        :type record_handler: RecordDAO
        :param protected: list of protected parameters, e.g internal params not to be stored
        :type protected: list, optional
        :param metadata: dictionary of attributes/value to initialize object with, defaults to {}
        :type metadata: dict, optional
        :param record: sina record to prevent looking it up again and again in sina
        :type record: Record
        """
        with store.lock_strategy:
            self.__dict__["__store__"] = store
            self.__dict__["__schema__"] = schema
            self.__dict__["__record_handler__"] = record_handler
            self.__dict__["__protected__"] = [
                "id", "__type__", "__protected__",
                "__record_handler__", "__store__", "id", "__schema__"] + protected
            self.__dict__["__type__"] = kosh_type
            self.__dict__["lock_strategy"] = store.lock_strategy
            if Id is None:
                Id = uuid.uuid4().hex
                record = Record(id=Id, type=kosh_type, user_defined={'kosh_information': {}})
                if store.__sync__:
                    store.lock()
                    store.__record_handler__.insert(record)
                    store.unlock()
                else:
                    record["user_defined"]['kosh_information']["last_update_from_db"] = time.time()
                    self.__store__.__sync__dict__[Id] = record
                self.__dict__["id"] = Id
            else:
                self.__dict__["id"] = Id
                if record is None:
                    try:
                        record = self.get_record()
                    except BaseException:  # record exists nowhere
                        record = Record(id=Id, type=kosh_type, user_defined={'kosh_information': {}})
                        if store.__sync__:
                            store.lock()
                            store.__record_handler__.insert(record)
                            store.unlock()
                        else:
                            self.__store__.__sync__dict__[Id] = record
                            record["user_defined"]['kosh_information']["last_update_from_db"] = time.time()
                else:
                    deleted_items = False
                    for att in self.__dict__["__protected__"]:
                        if att in record["data"]:
                            del record["data"][att]
                            deleted_items = True
                    if deleted_items:
                        if store.__sync__:
                            self._update_record(record)
                        else:
                            self.__store__.__sync__dict__[Id] = record

            metadata_copy = metadata.copy()
            for key in metadata:
                if key in self.__dict__["__protected__"]:
                    del metadata_copy[key]
            self.update(metadata_copy)

    @lock_strategies.lock_method
    def __getattr__(self, name):
        """__getattr__ get an attribute

        :param name: attribute to retrieve
        :type name: str
        :raises AttributeError: could not retrieve attribute
        :return: requested attribute value
        """
        if name == "__id__":
            warnings.warn(
                "the attribute '__id__' has been deprecated in favor of 'id'",
                DeprecationWarning)
            name = "id"
        if name in self.__dict__["__protected__"]:
            if name == "_associated_datasets_" and self.__type__ == self.__store__._ensembles_type:
                rels = self.get_sina_store().relationships.find(
                    None, "is a member of ensemble", self.id)
                return [str(x.subject_id) for x in rels]
            if name == "__features__":
                record = self.get_record()
                try:
                    return KoshPickler().loads(record["user_defined"]["__features__"])
                except Exception:
                    return {None: {}}
            if name == "_associated_data_":
                from kosh.dataset import KoshDataset
                record = self.get_record()
                # Any curve sets?
                if len(record["curve_sets"]) != 0:
                    out = [self.id, ]
                else:
                    out = []
                # we cannot use list comprehension
                # some pure sina rec have file but no kosh_id
                for file_rec in record["files"]:
                    file_entry = record["files"][file_rec]
                    if "kosh_id" in file_entry:
                        out.append(record["files"][file_rec]["kosh_id"])
                    else:
                        # Not an entry made by Kosh
                        # But maybe we can salvage this!
                        # did  the user added a mime_type?
                        if "mimetype" in file_entry:
                            out.append("{}__uri__{}".format(self.id, file_rec))
                # Now we need to add the parent ensembles associated data
                if isinstance(self, KoshDataset):
                    for ensemble in self.get_ensembles():
                        out += ensemble._associated_data_
                return out
            else:
                return self.__dict__[name]
        record = self.get_record()
        if name == "__attributes__":
            return self.__getattributes__()
        elif name == "schema":
            if self.__dict__[
                    "__schema__"] is None and "schema" in record["data"]:
                schema = kosh_pickler.loads(record["data"]["schema"]["value"])
                self.__dict__["__schema__"] = schema
            return self.__dict__["__schema__"]
        elif name in ['alias_feature', 'loader_kwargs']:
            if name in record["data"]:
                return kosh_pickler.loads(record["data"][name]["value"])
            else:
                return {}
        if name not in record["data"]:
            if name == "mime_type":
                return record["type"]
            elif name == "uri":
                return ""
            else:
                raise AttributeError(
                    "Object {} does not have {} attribute".format(self.id,
                                                                  name))
        value = record["data"][name]["value"]
        if name == "creator":
            # old records have user id let's fix this
            if value in self.__store__.__record_handler__.find_with_type(
                    self.__store__._users_type, ids_only=True):
                value = self.__store__.get_record(value)["data"]["username"]["value"]
        return value

    @lock_strategies.lock_method
    def get_sina_store(self):
        """Returns the sina store object"""
        return self.__store__.get_sina_store()

    @lock_strategies.lock_method
    def get_sina_records(self):
        """Returns sina store's records"""
        return self.__record_handler__

    @lock_strategies.lock_method
    def update(self, attributes):
        """update many attributes at once to limit db writes
        :param: attributes: dictionary with attributes to update
        :type attributes: dict
        """
        if 'id' in attributes:
            del attributes['id']
        rec = None
        N = len(attributes)
        n = 0
        for name, value in attributes.items():
            n += 1
            if n == N:
                update_db = True
            else:
                update_db = False
            rec = self.___setattr___(name, value, rec, update_db=update_db)

    @lock_strategies.lock_method
    def __setattr__(self, name, value):
        """set an attribute
        We are calling the ___setattr___
        because of special case that needs extra args and return values
        """
        self.___setattr___(name, value)

    @lock_strategies.lock_method
    def ___setattr___(self, name, value, record=None,
                      update_db=True, force=False):
        """__setattr__ set an attribute on an object

        :param name: name of attribute
        :type name: str
        :param value: value to set attribute to
        :type value: object
        :param record: sina record if already extracted before, save db access
        :type record: sina.model.Record
        :param force: force dataset attribute setting (when sent from ensemble)
        :type force: bool
        :return: sina record updated
        :rtype: sina.model.Record
        """
        if name in self.__protected__:  # Cannot set protected attributes
            return record
        if record is None:
            record = self.get_record()
        if name == "schema":
            assert isinstance(value, KoshSchema)
            value.validate(self)
        elif name == "__features__":
            value = kosh_pickler.dumps(value)
            record["user_defined"]["__features__"] = value
        elif name in ['alias_feature', 'loader_kwargs']:
            if isinstance(value, dict):
                value = kosh_pickler.dumps(value)
            else:  # Pre-pickled at dataset level
                value = value
        elif self.schema is not None:
            self.schema.validate_attribute(name, value)

        # For datasets we need to check if the att comes from ensemble
        from kosh.dataset import KoshDataset
        if isinstance(self, KoshDataset) and not force:
            sina_store = self.get_sina_store()
            # Let's get the relationships it's in
            relationships = sina_store.relationships.find(
                self.id, self.__store__._ensemble_predicate, None)
            for relationship in relationships:
                ensemble = self.__store__.open(relationship.object_id)
                ens_tags = self.list_ensemble_tags(ensemble.id, dictionary=True, obscure=False)
                inherit_attributes = ens_tags.get(f"{ensemble.id}_ENSEMBLE_TAG_INHERIT_ATTRIBUTES", True)
                if inherit_attributes:
                    if name in ensemble.list_attributes() and name not in ensemble.__dict__["__ok_duplicates__"]:
                        if value != getattr(ensemble, name):
                            raise KeyError(
                                "The attribute {} is controlled by ensemble: {} and cannot be set here".format(
                                    name, relationship.object_id))
                        else:
                            warnings.warn(
                                "The attribute {} is controlled by ensemble: {}"
                                ". You should NOT set this attribute at the dataset level"
                                ". Values match so we will accept it here".format(
                                    name, relationship.object_id), UserWarning)

        # For Ensembles we need to set it on all members
        from kosh.ensemble import KoshEnsemble
        if isinstance(self, KoshEnsemble):
            # First we make a pass to collect all other ensembles datasets are
            # part of
            for dataset in self.get_members():
                for ensemble in dataset.get_ensembles():
                    ens_tags = dataset.list_ensemble_tags(ensemble.id, dictionary=True, obscure=False)
                    inherit_attributes = ens_tags.get(f"{ensemble.id}_ENSEMBLE_TAG_INHERIT_ATTRIBUTES", True)
                    if inherit_attributes:
                        if ensemble.id != self.id:
                            for att in ensemble.list_attributes():
                                if att in self.__dict__["__ok_duplicates__"]:
                                    continue
                                if att == name:
                                    raise NameError("A member of this ensemble belongs to ensemble {} "
                                                    "which already controls attribute {}".format(ensemble.id, att))

                        dataset.___setattr___(
                            name=name,
                            value=value,
                            record=None,
                            update_db=update_db,
                            force=True)

        # Did it change on db since we last read it?
        last_modif_att = "{name}_last_modified".format(name=name)
        try:
            # Time we last read its value
            last = self.__dict__[last_modif_att]
        except KeyError:
            last = time.time()
        try:
            # Time we last read its value
            last_db = record["user_defined"]['kosh_information'][last_modif_att]
        except KeyError:
            last_db = last
        # last time attribute was modified in db
        if last_db > last and getattr(
                self, name) != record["data"][name]["value"]:  # Ooopsie someone touched it!
            raise AttributeError("Attribute {} of object id {} was modified since last sync\n"
                                 "Last modified in db at: {}, value: {}\n"
                                 "You last read it at: {}, with value: {}".format(
                                     name, self.id,
                                     last_db, record["data"][name],
                                     last, getattr(self, name)))
        now = time.time()
        if "{name}_last_modified".format(name=name) not in self.__protected__:
            self.__dict__["__protected__"] += [last_modif_att, ]
        self.__dict__[last_modif_att] = now
        record["user_defined"]['kosh_information'][last_modif_att] = now
        if name == "schema":
            self.__dict__["__schema__"] = value
            value = kosh_pickler.dumps(value)
        record["data"][name] = {"value": value}
        if update_db and self.__store__.__sync__:
            self._update_record(record)
        return record

    @lock_strategies.lock_method
    def _update_record(self, record, store=None):
        """Updates a record in the sina store
        :param record: The record to update
        :type record: sina.model.Record
        :param store: sina store to update
        :type store: sina.datastore.DataStore"""
        self.__store__.lock()
        if store is None:
            store = self.__store__.get_sina_store()
        id_ = record.id
        rels = store.relationships.find(id_, None, None)
        rels += store.relationships.find(None, None, id_)
        if not self.__store__.__sync__:
            store.records.delete(id_)
            store.records.insert(record)
            store.relationships.insert(rels)
        store.records.update(record)
        self.__store__.unlock()

    @lock_strategies.lock_method
    def __delattr__(self, name):
        """__delattr__ deletes an attribute

        :param name: attribute to delete
        :type name: str
        """
        if name in self.__protected__:
            return
        record = self.get_record()
        last_modif_att = "{name}_last_modified".format(name=name)
        now = time.time()
        record["user_defined"]['kosh_information'][last_modif_att] = now
        # We need to remember we touched it otherwise
        # if we create it again the db will look
        # out of sync.
        self.__dict__[last_modif_att] = now
        del record["data"][name]
        if self.__store__.__sync__:
            self._update_record(record)

    @lock_strategies.lock_method
    def sync(self):
        """sync this object with database"""
        self.__store__.sync([self.id, ])

    @lock_strategies.lock_method
    def list_attributes(self, dictionary=False, ensemble_id=None):
        __doc__ = self.listattributes.__doc__.replace("listattributes", "list_attributes")  # noqa
        return self.listattributes(dictionary=dictionary, ensemble_id=ensemble_id)

    @lock_strategies.lock_method
    def listattributes(self, dictionary=False, ensemble_id=None):
        """listattributes list all non protected attributes

        :param dictionary: return a dictionary of value/pair rather than just attributes names
        :type dictionary: bool
        :param ensemble_id: Provide ensemble ID(s) to return ensemble tags
        :type ensemble_id: str, lst

        :return: list of attributes set on object
        :rtype: list
        """
        record = self.get_record()
        attributes = list(record["data"].keys()) + ['id', ]

        no_tags = [a for a in attributes if "_ENSEMBLE_TAG_" not in a]  # Remove ensemble tags
        ens_tags = []
        if ensemble_id is not None:
            if isinstance(ensemble_id, str):
                ensemble_id = [ensemble_id]
            for ens_id in ensemble_id:
                ens_tags += sorted([a for a in attributes if f"{ens_id}_ENSEMBLE_TAG_" in a])
        attributes = sorted(no_tags)

        for att in self.__protected__:
            if att in attributes and att != "id":
                attributes.remove(att)
        attributes += ens_tags
        if dictionary:
            out = {}
            for att in attributes:
                out[att] = getattr(self, att)
            return out
        else:
            return attributes

    @lock_strategies.lock_method
    def __getattributes__(self):
        """__getattributes__ return dictionary with pairs of attribute/value

        :return: dictionary with pairs of attribute/value
        :rtype: dict
        """
        record = self.get_record()
        attributes = {}
        for a in record["data"]:
            if a in ['alias_feature', 'loader_kwargs']:
                continue
            attributes[a] = record["data"][a]["value"]
            if a == "creator":
                # old records have user id let's fix this
                if attributes[a] in self.__store__.__record_handler__.find_with_type(
                        self.__store__._users_type, ids_only=True):
                    attributes[a] = self.__store__.get_record(attributes[a])["data"]["username"]["value"]
        return attributes

    @lock_strategies.lock_method
    def __str__(self):
        """String for printing"""
        if self.__dict__["__store__"].verbose_attributes:
            def reprtool(item):
                return item
        else:
            def reprtool(item):
                if isinstance(item, str):
                    return reprlib.repr(item)[1:-1]
                else:
                    return reprlib.repr(item)
        st = "Id: {}".format(self.id)
        for att in sorted(self.listattributes()):
            if att != 'id':
                st += "\n\t{}: {}".format(att, reprtool(getattr(self, att)))
        return st


class KoshSinaFile(KoshSinaObject):
    """KoshSinaFile file representation in Kosh via Sina"""

    @lock_strategies.lock_method
    def open(self, *args, **kargs):
        """open opens the file
        :return: handle to file in open mode
        """
        return self.__store__.open(self.id, *args, **kargs)



# this defines Kosh transformers object to augment/filter data from loaders
from abc import ABCMeta, abstractmethod
import hashlib
import os
from ..exec_graphs import KoshExecutionGraph
from kosh import kosh_cache_dir


class KoshTransformer(KoshExecutionGraph):
    # Defines which input types it can handle
    # and what output it sends back
    __metaclass__ = ABCMeta
    types = {"numpy": ["numpy", ]}

    def __init__(self,
                 cache_dir=kosh_cache_dir,
                 cache=False, verbose=False, *args, **kargs):
        """init function will receive the previous step's signature and the cache directory
        and output signature is also generated from the input args (w/o the cache_dir)
        :param cache_dir: directory to save cachd files
        :type cache_dir: str
        :param cache: do we use cache? 0: no, 1:yes, 2:yes but clobber if exists
        :type cache: int
        :param verbose: Turn on verbosity, by default this will turn on printing a message
                        when results are loaded from cache. Message is sent as lone argument
                        to `self._print` function.
                        value is stored in self._verbose
        :type verbose: bool
        """
        self.signature = hashlib.sha256(repr(self.__class__).encode()).hexdigest()
        self.signature = self.update_signature(*args, **kargs)
        self.cache_dir = cache_dir
        self._verbose = verbose
        if cache:
            try:
                os.makedirs(self.cache_dir)
            except Exception:
                pass
        self.cache = cache

    def transform_(self, input, format, signature=None):
        """Given input from previous loader or transformer and desired format
        computes the unique signature and tries to extract from cache, calls transformer's
        `transform` function if no cache available.
        :param input: set of input passed from loader or previous transformer
        :type input: object
        :param format: desired output format
        :type format: str
        :return: The result from transform function
        :rtype: object
        """

        if self.cache:
            if signature is None:
                use_signature = self.update_signature(
                    input, format)
            else:
                use_signature = signature

            cache_file = os.path.join(self.cache_dir, use_signature)
            if self.cache == 2 and os.path.exists(cache_file):
                # User wants to clobber cache
                os.remove(cache_file)

            try:
                result = self.load(use_signature)
                if self._verbose:
                    self._print("Loaded results from cache file {} using signature: {}".format(
                        cache_file, use_signature))
            except Exception:
                if signature is None:
                    signature = self.update_signature(
                        input, format)
                result = self.transform(input, format)
                if self.cache > 0:  # Ok user wants to cache results
                    if not os.path.exists(self.cache_dir):
                        os.makedirs(self.cache_dir)
                    self.save(signature, result)
        else:
            result = self.transform(input, format)
        return result

    def _print(self, message):
        print(message)

    @abstractmethod
    def transform(self, input_, format):
        """The transform function
        :param input_: result returned by loader or previous transformer
        """
        raise NotImplementedError("the transform function is not implemented")


# Convenience decorators
def typed_transformer(transformer_types=None):
    if transformer_types is None:
        transformer_types = {"numpy": ["numpy", ]}

    def actual_transformer(func):
        class CustomTypedTransformer(KoshTransformer):
            types = transformer_types

            def transform(self, inputs, format):
                return func(inputs)

            def __getitem_propagate__(self, key, input_index):
                return key

        return CustomTypedTransformer()
    return actual_transformer


def typed_transformer_with_format(transformer_types={"numpy": ["numpy", ]}):
    def actual_transformer(func):
        class CustomTypedTransformer(KoshTransformer):
            types = transformer_types

            def transform(self, inputs, format):
                return func(inputs, format)

            def __getitem_propagate__(self, key, input_index):
                return key
        return CustomTypedTransformer()
    return actual_transformer


numpy_transformer = typed_transformer()



import warnings
from .dataset import KoshDataset
from .utils import cleanup_sina_record_from_kosh_sync
from .utils import update_json_file_with_records_and_relationships
from .utils import __check_valid_connection_type__
from . import lock_strategies
try:
    import orjson
except ImportError:
    import json as orjson  # noqa
import kosh


class KoshEnsemble(KoshDataset):
    def __init__(self, id, store, schema=None, record=None):
        """Kosh Ensemble
Ensemble allows to link together many datasets.
These datasets will inherit attributes and associated sources from the ensemble.

        :param id: dataset's unique Id
        :type id: str
        :param store: store containing the dataset
        :type store: KoshSinaStore
        :param schema: Kosh schema validator
        :type schema: KoshSchema
        :param record: to avoid looking up in sina pass sina record
        :type record: Record
        """
        with store.lock_strategy:
            super(KoshEnsemble, self).__init__(id, store,
                                               schema=schema, record=record,
                                               kosh_type=store._ensembles_type)
            self.__dict__["__protected__"] = ["__name__", "__creator__", "__store__",
                                              "_associated_data_", "__features__",
                                              "_associated_datasets_", "__ok_duplicates__"]
            # Attributes that the members can have on their own
            self.__dict__["__ok_duplicates__"] = ["creator", "id", "name"]

    @lock_strategies.lock_method
    def __str__(self):
        """string representation"""
        st = super(KoshEnsemble, self).__str__()
        st = st.replace("KOSH DATASET", "KOSH ENSEMBLE")
        st = st[:st.find("--- Ensembles") - 1]
        if self._associated_datasets_ is not None:
            st += "\n--- Member Datasets ({})---\n".format(
                len(self._associated_datasets_))
            st += "\t{}".format(self._associated_datasets_)
        return st

    @lock_strategies.lock_method
    def cleanup_files(self, dry_run=False, interactive=False, **search_keys):
        """Cleanup the ensemble's members from references to dead files.
        You can filter associated objects by passing key=values
        e.g mime_type=hdf5 will only dissociate non-existing files associated with mime_type hdf5
        some_att=some_val will only dissociate non-existing files associated and having the attribute
        'some_att' with value of 'some_val'
        returns list of uris to be removed.
        :param dry_run: Only does a dry_run
        :type dry_run: bool
        :param interactive: interactive mode, ask before dissociating
        :type interactive: bool
        :returns: list of uris (to be) removed.
        :rtype: list
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        missings = super(
            KoshEnsemble,
            self).cleanup_files(
            dry_run=dry_run,
            interactive=interactive,
            **search_keys)
        for dataset in self.get_members():
            missings += dataset.cleanup_files(dry_run=dry_run,
                                              interactive=interactive, **search_keys)
        return missings

    @lock_strategies.lock_method
    def export(self, file=None):
        """Exports this ensemble datasets
        :param file: export datasets to a file
        :type file: None or str
        :return: dataset and its associated data
        :rtype: dict"""
        records = [cleanup_sina_record_from_kosh_sync(self.get_record()), ]
        for dataset_id in self.get_members(ids_only=True):
            records.append(
                cleanup_sina_record_from_kosh_sync(
                    self.__store__.get_record(dataset_id)))
        # We also need to export the relationships
        relationships = self.get_sina_store().relationships.find(
            None, self.__store__._ensemble_predicate, self.id)
        output_dict = {
            "minimum_kosh_version": None,
            "kosh_version": kosh.version(comparable=True),
            "sources_type": self.__store__._sources_type,
            "records": records,
            "relationships": relationships
        }

        update_json_file_with_records_and_relationships(file, output_dict)
        return output_dict

    @lock_strategies.lock_method
    def create(self, name="Unnamed Dataset", id=None,
               metadata={}, schema=None, sina_type=None, ensemble_tags=None,
               inherit_attributes=True, **kargs):
        """create a new (possibly named) dataset as a member of this ensemble.

        :param name: name for the dataset, defaults to None
        :type name: str, optional
        :param id: unique Id, defaults to None which means use uuid4()
        :type id: str, optional
        :param metadata: dictionary of attribute/value pair for the dataset, defaults to {}
        :type metadata: dict, optional
        :param schema: a KoshSchema object to validate datasets and when setting attributes
        :type schema: KoshSchema
        :param sina_type: If you want to query the store for a specific sina record type, not just a dataset
        :type sina_type: str
        :param inherit_attributes: Whether datasets inherit attributes from ensembles.
                                   If False, datasets can have the same attributes as ensembles
                                   and the different ensembles that the dataset belongs to can also
                                   have the same attributes. They can also have different values.
                                   Defaults to True.
        :type inherit_attributes: bool
        :param ensemble_tags: Organize datasets within ensemble by using tags. These "attributes"
                              are only available at the ensemble level for each dataset. These tags
                              can also be used in `ensemble.find_datasets(ensemble_tags=dict)`.
                              e.g., ensemble_tags={"even_or_odd": "even", "data_type": "test data"}
        :type ensemble_tags: dict
        :param kargs: extra keyword arguments (ignored)
        :type kargs: dict
        :raises RuntimeError: Dataset already exists
        :return: KoshDataset
        :rtype: KoshDataset
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        if sina_type == self.__store__._ensembles_type:
            raise ValueError("You cannot create an ensemble from an ensemble")

        attributes = self.list_attributes()
        for key in metadata:
            if key in attributes:
                if metadata[key] != getattr(self, key):
                    raise ValueError(
                        "'{}' is an attribute of this ensemble and "
                        "therefore cannot be an attribute of its descendants".format(key))
                else:
                    warnings.warn(
                        "'{}' is an attribute of this ensemble and "
                        "therefore cannot be an attribute of its descendants"
                        ". Values match so we will accept it here.".format(key), UserWarning)

        ds = self.__store__.create(
            name=name,
            id=id,
            metadata=metadata,
            schema=schema,
            sina_type=sina_type,
            **kargs)
        self.add(ds, inherit_attributes=inherit_attributes, ensemble_tags=ensemble_tags)
        return ds

    @lock_strategies.lock_method
    def add(self, dataset, inherit_attributes=True, ensemble_tags=None):
        """Adds a dataset to this ensemble
        :param dataset: The dataset to add to this ensemble
        :type dataset: KoshDataset or str
        :param inherit_attributes: Whether datasets inherit attributes from ensembles.
                                   If False, datasets can have the same attributes as ensembles
                                   and the different ensembles that the dataset belongs to can also
                                   have the same attributes. They can also have different values.
                                   Defaults to True.
        :type inherit_attributes: bool
        :param ensemble_tags: Organize datasets within ensemble by using tags. These "attributes"
                              are only available at the ensemble level for each dataset. These tags
                              can also be used in `ensemble.find_datasets(ensemble_tags=dict)`.
                              e.g., ensemble_tags={"even_or_odd": "even", "data_type": "test data"}
        :type ensemble_tags: dict
        """
        from sina.model import Relationship
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        # Step1 make sure the dataset does not belong to another ensemble
        if isinstance(dataset, KoshDataset):
            dataset_id = dataset.id
        else:
            dataset_id = dataset
            dataset = self.__store__._load(dataset_id)
        relationships = self.get_sina_store().relationships.find(
            dataset_id, self.__store__._ensemble_predicate, None)

        if inherit_attributes:
            for rel in relationships:
                if rel.object_id != self.id:
                    other_ensemble = self.__store__.open(rel.object_id)
                    # Ok... Already a member of another ensemble.
                    # let's make sure there are no conflict here
                    for att in self.list_attributes():
                        if att in self.__dict__["__ok_duplicates__"]:
                            continue
                        if att in other_ensemble.list_attributes():
                            raise ValueError(
                                "Dataset {} is already part of ensemble {} "
                                "which already provides support for attribute: {}. Bailing".format(
                                    dataset_id, rel.object_id, att))
                else:
                    # ok it's already done, no need to do anything else
                    return

            # Ok we're good, let's now makes sure attributes are ok
            attributes = self.list_attributes(dictionary=True)
            dataset_attributes = dataset.list_attributes(dictionary=True)
            for att in dataset.list_attributes():
                if att in self.__dict__["__ok_duplicates__"]:
                    continue
                if att in attributes and dataset_attributes[att] != attributes[att]:
                    raise ValueError(
                        f"Dataset {dataset_id} has attribute `{att}` with value {dataset_attributes[att]}, "
                        f"this ensemble ({self.id}) has value `{attributes[att]}`\n"
                        "This error can by bypassed by setting `inherit_attributes=False` which makes it so "
                        "datasets can have the same attributes as ensembles and the different ensembles that "
                        "the dataset belongs to can also have the same attributes.")
            # At this point we need to add the ensemble attributes to the dataset
            for att in self.list_attributes():
                if att in self.__dict__["__ok_duplicates__"]:
                    continue
                dataset.___setattr___(att, getattr(self, att), force=True)
        else:
            # Use for core_sina.py ___setattr___()
            if isinstance(ensemble_tags, dict):
                ensemble_tags['INHERIT_ATTRIBUTES'] = False
            else:
                ensemble_tags = {'INHERIT_ATTRIBUTES': False}
        # Ok We are clear let's create the relationship
        rel = Relationship(
            self.id, dataset_id, self.__store__._ensemble_predicate)
        # Add ensemble tags
        if ensemble_tags is not None:
            if self.schema is not None:
                self.schema.validate(ensemble_tags, f"Ensemble {self.id}")
            if dataset.schema is not None:
                dataset.schema.validate(ensemble_tags, f"Dataset {dataset.id}")
            dataset.add_ensemble_tags(self.id, ensemble_tags)
        self.get_sina_store().relationships.insert(rel)

    @lock_strategies.lock_method
    def remove(self, dataset):
        """Removes a dataset from this ensemble. Does not delete the dataset.
        :param dataset: The dataset to remove
        :type dataset: KoshDataset or str
        """
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write'])
        # Step1 make sure the dataset does not belong to another ensemble
        if isinstance(dataset, KoshDataset):
            dataset_id = dataset.id
        else:
            dataset_id = dataset
            dataset = self.__store__.open(dataset_id)
        relationships = self.get_sina_store().relationships.find(
            dataset_id, self.__store__._ensemble_predicate, self.id)
        # Delete ensemble tags from dataset
        ensemble_tags = dataset.list_ensemble_tags(ensemble_id=self.id)
        ensemble_tags = [et.replace(f"{self.id}_ENSEMBLE_TAG_", "") for et in ensemble_tags]
        dataset.delete_ensemble_tags(ensemble_id=self.id, ensemble_tags=ensemble_tags)
        if len(relationships) == 0:
            warnings.warn(
                "Dataset {} is not a member of ensemble {}".format(
                    dataset_id, self.id))
            return

        rel = relationships[0]
        self.get_sina_store().relationships.delete(rel.subject_id, rel.predicate, rel.object_id)

    delete = remove

    @lock_strategies.lock_method
    def get_members(self, ids_only=False):
        """Generator for member datasets
        :param ids_only: generator will return ids if True Kosh datasets otherwise
        :type ids_only: bool
        :returns: generator of dataset (or ids)
        :rtype: str or KoshDataset
        """
        for id in self._associated_datasets_:
            if ids_only:
                yield id
            else:
                yield self.__store__.open(id)

    @lock_strategies.lock_method
    def find_datasets(self, ensemble_tags=None, *atts, **keys):
        """Find datasets members of this ensemble that are matching some metadata.
        Arguments are the metadata names we are looking for e.g
        find("attr1", "attr2")
        you can further restrict by specifying exact value for a metadata
        via key=value
        you can return ids only by using: ids_only=True
        range can be specified via: sina.utils.DataRange(min, max)

        :param ensemble_tags: Further filter out datasets with ensemble tags
        :type ensemble_tags: dict
        :return: generator of matching datasets in this ensemble
        :rtype: generator
        """
        if ensemble_tags is not None:
            for key, val in ensemble_tags.items():
                keys[f"{self.id}_ENSEMBLE_TAG_{key}"] = val
        members_ids = list(self.get_members(ids_only=True))
        return self.__store__.find(id_pool=members_ids, *atts, **keys)

    @lock_strategies.lock_method
    def clone(self, *atts, **keys):
        """We cannot clone an ensemble"""
        __check_valid_connection_type__(self.__store__.__connection_type__, ['write', 'append'])
        raise NotImplementedError("Ensembles objects cannot clone themselves")

    @lock_strategies.lock_method
    def list_attributes(self, dictionary=False, no_duplicate=False):
        """list_attributes list all non protected attributes

        :param dictionary: return a dictionary of value/pair rather than just attributes names
        :type dictionary: bool

        :param no_duplicate: return only attributes that cannot be duplicated in members
        :type no_duplicate: bool

        :return: list of attributes set on object
        :rtype: list
        """

        attributes = super(KoshEnsemble, self).list_attributes(dictionary)
        if no_duplicate:
            return [x for x in attributes if x not in self.__dict__["__ok_duplicates__"]]
        else:
            return attributes

    @lock_strategies.lock_method
    def to_dataframe(self, data_columns=[], include_ensemble_attributes=True, include_ensemble_tags=True,
                     *atts, **keys):
        """Return the find_datasets object as a Pandas DataFrame.

        Pass in the same arguments and keyword arguments as the find method.

        Arguments are the metadata name we are looking for e.g
        find("attr1", "attr2")
        you can further restrict by specifying exact value for a metadata
        via key=value
        you can return ids only by using: ids_only=True
        range can be specified via: sina.utils.DataRange(min, max)

        "file_uri" is a reserved key that will return all records being associated
                   with the given "uri", e.g store.find(file_uri=uri)
        "types" let you search over specific sina record types only.
        "id_pool" will search based on id of Sina record or Kosh dataset. Can be a list.

        :param data_columns: Columns to extract. By default this will include ['id', 'name', 'creator'].
                             If nothing is passed, will return all data.
        :type data_columns: Union(str, list), optional
        :param include_ensemble_attributes: Include ensemble attributes in DataFrame.
        :type include_ensemble_attributes: bool, optional
        :param include_ensemble_tags: Include ensemble tags in DataFrame.
        :type include_ensemble_tags: bool, optional
        :return: Pandas DataFrame
        :rtype: Pandas DataFrame
        """
        import pandas as pd
        if isinstance(data_columns, str):
            data_columns = [data_columns]

        keys['load_type'] = 'dictionary'
        keys['ids_only'] = False
        datasets = list(self.find_datasets(*atts, **keys))

        attr_dict = {}
        total_datasets = len(datasets)

        # Always have these by default
        defaults = ['id', 'name', 'creator']

        unique_keys = set()
        for dataset in datasets:
            unique_keys.update(list(dataset['data'].keys()))
        data_columns_all = sorted(unique_keys)
        # Remove all ensemble tags and only keep dataset attributes
        data_columns_no_ensemble_tags = [dc for dc in data_columns_all if "_ENSEMBLE_TAG_" not in dc]

        # Acquire all data if `data_columns` was not passed
        if not data_columns:
            data_columns = data_columns_no_ensemble_tags

        data_columns = defaults + data_columns  # Want defaults in front

        # Acquire ensemble attributes
        if include_ensemble_attributes:
            data_columns_ens_default_attributes = [f"{self.id}_ENSEMBLE_ATTRIBUTE_id",
                                                   f"{self.id}_ENSEMBLE_ATTRIBUTE_name",
                                                   f"{self.id}_ENSEMBLE_ATTRIBUTE_creator"]

            ens_attrs = self.list_attributes(dictionary=True)
            data_columns_ens_other_attributes = [f"{self.id}_ENSEMBLE_ATTRIBUTE_{attr}" for attr in ens_attrs.keys()
                                                 if attr not in ['id', 'name', 'creator']]

            data_columns += data_columns_ens_default_attributes + data_columns_ens_other_attributes

        # Remove all other ensemble tags and only keep this ensemble's tags
        if include_ensemble_tags:
            data_columns += [dc for dc in data_columns_all if f"{self.id}_ENSEMBLE_TAG_" in dc and
                             "_ENSEMBLE_TAG_INHERIT_ATTRIBUTES" not in dc]

        attr_dict = {d: [pd.NA] * total_datasets for d in data_columns}

        for i, dataset in enumerate(datasets):
            attr_dict['id'][i] = dataset['id']
            for column, values in dataset['data'].items():
                if column in data_columns:
                    attr_dict[column][i] = values.get('value', pd.NA)

        # Ensemble attributes
        if include_ensemble_attributes:
            for key, val in ens_attrs.items():
                attr_dict[f"{self.id}_ENSEMBLE_ATTRIBUTE_{key}"] = [val] * total_datasets

        df = pd.DataFrame(attr_dict)
        return df



from __future__ import print_function, division
from .core import KoshLoader


def walk_hdf5(dataset, prefix=""):
    """Walk through hdf5 groups to find all datasets and return their paths
    return generator
    :param dataset: hdf5 dataset to start walking from
    :type dataset: h5py._hl.dataset.Dataset
    :param prefix: prefix to use when walking hdf5 paths
    :return: hdf5 dataset structure
    :rtype: generator
    """
    import h5py
    for key in sorted(dataset.keys()):
        value = dataset[key]
        if isinstance(value, h5py._hl.dataset.Dataset):
            yield prefix+"/"+key+"***"
        else:
            if prefix == "":
                yield "/".join(walk_hdf5(value, prefix=key))
            else:
                yield "/".join(walk_hdf5(value, prefix=prefix+"/"+key))


def list_hdf5(dataset):
    """walk hdf5 and return list of path to all datasets
    :param dataset: hdf5 dataset to start walking from
    :type dataset: h5py._hl.dataset.Dataset
    :return: hdf5 dataset structure
    :rtype: list
    """
    nest = list(walk_hdf5(dataset))
    out = []
    for p in nest:
        for d in p.split("***"):
            if len(d) > 0:
                # Removes leading /
                while d[0] == "/":
                    d = d[1:]
                out.append(d)
    return out


class HDF5Loader(KoshLoader):
    """ Kosh loader to load HDF5 data"""
    types = {"hdf5": ["numpy", ]}

    def open(self, mode='r'):
        """open/load  matching Kosh Sina File

        :param mode: mode to open the file in, defaults to 'r'
        :type mode: str, optional
        :return: Kosh File object
        """
        import h5py
        return h5py.File(self.obj.uri, mode)

    def extract(self):
        """extract return a feature from the loaded object.

        :param feature: variable to read from file
        :type feature: str
        :param format: desired output format
        :type format: str
        :return: data
        """
        import h5py
        args, kargs = self._user_passed_parameters
        f = h5py.File(self.uri, "r")
        features = self.feature
        if not isinstance(features, list):
            features = [self.feature, ]

        out = []
        for feature in features:
            feat = f[feature]
            if len(kargs) != 0:  # probably requested dims
                feat_dims = [x.label for x in feat.dims]
                if feat_dims == [""]:
                    # Probably a dimension (cycle?)
                    feat_dims = [feature, ]
                user_dims = {}
                for k in list(kargs.keys()):
                    if k in feat_dims:
                        user_dims[k] = kargs.pop(k)
                select = {}
                for dim in user_dims:
                    user_selection = user_dims[dim]
                    if isinstance(user_selection, slice):
                        indices = user_selection
                    else:  # User passed a value or values
                        values = f[dim][:].tolist()
                        indices = [values.index(x) for x in user_selection]
                    select[feat_dims.index(dim)] = indices
                selectors = []
                for i in range(len(feat.shape)):
                    if i in select:
                        selectors.append(select[i])
                    else:
                        selectors.append(slice(0, None))
                feat = feat[tuple(selectors)]
            out.append(feat)
        if not isinstance(self.feature, list):
            return out[0]
        else:
            return out

    def list_features(self, group=None, **kargs):
        """list_features list features in file,
        for hdf5 you can pass extra argument to navigate groups.

        :return: list of features available in file
        :rtype: list
        """
        import h5py
        with h5py.File(self.uri, "r") as f:
            features = list_hdf5(f)
        if group is not None:
            feats = []
            for f in features:
                if f[:len(group)] == group:
                    feats.append(f[len(group)+1:])
            features = feats
        # At this point we need to allow to select a group itself
        feats = set()
        for feature in features:
            sp = feature.split("/")
            for i in range(len(sp)):
                feats.add("/".join(sp[:i+1]))
        features = sorted(feats)
        return features

    def describe_feature(self, feature):
        """describe a feature

        :param feature: feature (variable) to read, defaults to None
        :type feature: str, optional if loader does not require this
        :return: dictionary describing the feature
        :rtype: dict
        """
        import h5py
        features = self._list_features()
        if feature not in features:
            raise ValueError("feature {feature} is not available".format(feature=feature))

        info = {}
        with h5py.File(self.uri, "r") as f:
            feature = f[feature]
            info["size"] = feature.shape
            info["format"] = "hdf5"
            info["type"] = feature.dtype
            if hasattr(feature, "dims"):
                dims = []
                for d in feature.dims:
                    specs = {}
                    specs["name"] = d.label
                    try:
                        specs["first"] = f[d.label][0]
                        specs["last"] = f[d.label][-1]
                        specs["length"] = len(f[d.label])
                    except Exception:
                        pass
                    dims.append(specs)
                info["dimensions"] = dims
        return info



#!/bin/bash sbang
#!/usr/bin/env python
# This implements Kosh's CLI
from __future__ import print_function
import argparse
from kosh.utils import merge_datasets_handler  # noqa
import kosh
import sys
import shlex
from subprocess import Popen, PIPE
import os
import tempfile
import ast
import glob
import json
import six


def get_all_files(opts):
    """given a list of files, directory or pattern
    returns the list of files describe by these
    :param opts: list of files, directory or pattern
    :type opts: list
    :return: list of files
    :rtype: list
    """
    files = []
    for filename in opts:
        if os.path.isdir(filename):
            for root, dirs, filenames in os.walk(filename):
                for fnm in filenames:
                    files.append(os.path.join(root, fnm))
        elif "*" in filename or "?" in filename:
            files += glob.glob(filename)
        elif os.path.exists(filename):
            files.append(filename)
    return files


def find_files_from_list(uris):
    """given a uri/path walks this path/pattern to get sub dir/files
    :param uris: list of uris to scan (or list of lists)
    :type uris: list
    :return: List of files
    :rtype: list
    """
    # figure out targets
    new_uris = []
    for uri in uris:
        if isinstance(uri, list):
            new_uris += uri
        else:
            new_uris.append(uri)
    # figure out patterns and dirs
    pop = []
    for index, filename in enumerate(list(new_uris)):
        if "*" in filename or "?" in filename:
            # It's a pattern
            pop.append(index)
            new_uris += glob.glob(filename)
        elif os.path.isdir(filename):
            pop.append(index)
            for root, dirs, filenames in os.walk(filename):
                for fnm in filenames:
                    new_uris.append(os.path.join(root, fnm))

    # cleanup list
    for index in pop[::-1]:
        new_uris.pop(index)

    return new_uris


def core_parser(description,
                usage=None, prog=None):
    """
    Return the core parser with arguments common to all operations
    :param description: Description for the argparse parser
    :type description: str
    :param usage: Usage string
    :type usage: str
    :param prog: Name of the program for argparse to print
    :type prog: str
    :return: arparse parser
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=description,
        prog=prog,
        usage=usage,
        epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
    parser.add_argument("--store", "-s", required=True,
                        help="Kosh store to use")
    parser.add_argument("--dataset_record_type", "-d", default="dataset",
                        help="type used by sina db that Kosh will recognize as dataset")
    parser.add_argument("--version", action="store_true",
                        help="print version and exit")
    return parser


def parse_metadata(terms):
    """
    Parse metadata for Kosh  queries
    param=value / param>value, etc...
    :param terms: list of strings conatining name/operator/value
    :term terms: list of str
    :return: Dictionary with name as key and matching sina find object as value
    :rtype: dict
    """
    from sina.utils import DataRange
    metadata = {}
    for term in terms:
        found = False
        for operator in ["<=", ">=", "<", ">", "="]:
            sp = term.split(operator)
            if len(sp) != 2:
                continue
            found = True
            key = sp[0]
            try:
                # converts to int/float/DataRange if possible
                value = eval(sp[1])
            except Exception:
                value = sp[1]
            if operator == ">=":
                value = DataRange(min=value, min_inclusive=True)
            elif operator == ">":
                value = DataRange(min=value, min_inclusive=False)
            elif operator == "<=":
                value = DataRange(max=value, max_inclusive=True)
            elif operator == "<":
                value = DataRange(max=value, max_inclusive=False)
            metadata[key] = value
            break

        if not found:
            raise ValueError("Metadata must be in form 'key=value'")
    return metadata


def process_cmd(command, use_shell=False, shell="/usr/bin/bash"):
    """ Convenience function to run a command
    :param command: command to run
    :type command: str
    :param use_shell: ssh needs to be run as 'shell' and communicated command
                      This let us decide if it's our way to run the command
    :type use_shell: bool
    :param shell: If using a shell, this tells which shell to use
    :type shell: str
    :return: process object and output and error streams
    :rtype: list
    """

    if use_shell:
        if not sys.platform.startswith("win"):
            proc = Popen(shell, stdin=PIPE, stdout=PIPE, stderr=PIPE)
            o, e = proc.communicate(command.encode())
        else:
            proc = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
            o, e = proc.communicate()
            print("COMMAND:", command)
            print("OUT:", o.decode())
            print("ERR:", e.decode())
    else:
        if not sys.platform.startswith("win"):
            command = shlex.split(command)
        proc = Popen(command, stdout=PIPE, stderr=PIPE)
        o, e = proc.communicate()

    return proc, o, e


def open_stores(uris, dataset_record_type):
    """Given a list of paths/uri to Kosh stores this opens them all
    :param uris: list of Kosh stores to open
    :type uris: list
    :param dataset_record_type: record type in the store(s) or list of
    :type dataset_record_type: list or str
    :return: list of kosh stores
    :rtype: list
    """
    stores = []
    if isinstance(dataset_record_type, six.string_types):
        dataset_record_type = [dataset_record_type, ] * len(uris)
    for index, store_uri in enumerate(uris):
        if "@" in store_uri:
            # Remote store let's go fetch it
            _, store_local_uri = tempfile.mkstemp()
            cmd = "scp {} {}".format(store_uri, store_local_uri)
            p, o, e = process_cmd(cmd, use_shell=True)
            if p.returncode != 0:
                raise RuntimeError("Could not fetch remote store:", store_uri)
        else:
            store_local_uri = store_uri
        store = kosh.KoshStore(
            db_uri=store_local_uri,
            dataset_record_type=dataset_record_type[index])
        stores.append(store)
    return stores


def close_stores(stores, uris):
    """Closes a list of stores and if it was remote send it back to remote
    :param stores: List of Kosh store objects
    :type stores: list
    :param uris: list of Kosh stores to open
    :type uris: list
    """
    for i, store in enumerate(stores):
        # store.close()
        if store.db_uri != uris[i]:
            # Ok this one is a temporary file we need to send it back to remote
            cmd = "scp {} {}".format(store.db_uri, uris[i])
            p, o, e = process_cmd(cmd, use_shell=True)
            if p.returncode != 0:
                raise RuntimeError(
                    "Could not send updated store to {}".format(
                        uris[i]))


class KoshCmd(object):
    """Engine to dispatch kosh command to appropriate function"""
    def __init__(self):
        commands = "".join(
            ["" if k[0] == "_" else "\n\t" + k for k in sorted(dir(self))])

        # search is deprecated let's not list it
        index = commands.find("search")
        if index > -1:
            commands = commands[:index] + commands[index+8:]  # 8 because of \n\t

        parser = core_parser(
            description='Execute kosh operations',
            usage='''kosh <command> [<args>]

Available commands are:
    {commands}
'''.format(commands=commands))
        parser.add_argument('command', help='Subcommand to run')
        # first we parse all of it to catch --version
        help = False
        if (
                "--help" in sys.argv or "-h" in sys.argv) and sys.argv[1] in commands:
            help = True
            # Let's temporarily yank help from sys.argv
            try:
                help_index = sys.argv.index("-h")
            except Exception:
                pass
            try:
                help_index = sys.argv.index("--help")
            except Exception:
                pass
            sys.argv.pop(help_index)
        args, opts = parser.parse_known_args(sys.argv + ["-s", "blah"])
        if args.version or ("-v" in opts and "tar" not in opts):
            print("Kosh version:", kosh.__version__)
            sys.exit(0)
        # Ok now we parse only the rest to catch the command
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2] + ["-s", "blah"])
        if not hasattr(self, args.command) or args.command[0] == "_":
            print('Unrecognized command: {args.command}'.format(args=args))
            print('Known commands: {}'.format(" ".join(commands)))
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        if help:
            sys.argv.insert(help_index, "--help")
        getattr(self, args.command)()

    def search(self):
        """
        Deprecated use find
        """
        warnings.warn(DeprecationWarning, "The 'search' command is deprecated and now called `find`.\n"  # noqa: F821
                      "Please update your code to use `find` as `search` might disappear in the future")
        return self.find()

    def find(self):
        """find in a store command"""
        parser = core_parser(
            description='Find datasets in store that are matching metadata in form key=value')
        parser.add_argument(
            "--print",
            "-p",
            help="print each dataset info",
            action="store_true")
        args, search_terms = parser.parse_known_args(sys.argv[2:])
        metadata = parse_metadata(search_terms)
        store = kosh.KoshStore(db_uri=args.store,
                               dataset_record_type=args.dataset_record_type)
        metadata["ids_only"] = True
        ids = store.find(**metadata)
        if args.print:
            for Id in ids:
                ds = store.open(Id)
                print(ds)
                print(
                    "=======================================================================")
        else:
            print("\n".join(ids))

    def cleanup_files(self):
        """Cleanup a store from references to dead files
        You can filter associated object by matching metadata in form key=value
        e.g mime_type=hdf5 will only dissociate non-existing files associated with mime_type hdf5
        some_att=some_val will only dissociate non-existing files associated and having the attribute
        "some_att" with value of "some_val"""
        parser = core_parser(
            prog="kosh clean",
            description="""Cleanup a store from references to dead files
        You can filter associated object by matching metadata in form key=value
        e.g mime_type=hdf5 will only dissociate non-existing files associated with mime_type hdf5
        some_att=some_val will only dissociate non-existing files associated and having the attribute
        'some_att' with value of 'some_val'""")
        parser.add_argument("--dry-run", "--rehearsal", "-r", "-D", help="Dry run only, list ids of "
                            "dataset that would be cleaned up and path of files", action='store_true')
        parser.add_argument("--interactive", "-i", help="interactive mode, ask before dissociating",
                            action="store_true")
        args, search_terms = parser.parse_known_args(sys.argv[2:])
        metadata = parse_metadata(search_terms)
        store = kosh.KoshStore(db_uri=args.store,
                               dataset_record_type=args.dataset_record_type)
        ids = store.find(ids_only=True)
        for Id in ids:
            ds = store.open(Id)
            missings = ds.cleanup_files(dry_run=args.dry_run, interactive=args.interactive, **metadata)
            if len(missings) != 0:
                if not args.interactive:  # already printed in interactive
                    print(ds)
                for uri in missings:
                    associated = list(ds.find(uri=uri))
                    if len(associated) != 0:
                        print("{} (mime_type={}) is missing".format(
                            associated[0].uri, associated[0].mime_type))

    def create_ensemble(self):
        """create an ensemble into a Kosh store command"""
        parser = core_parser(
            prog="kosh create_ensemble",
            description='Creates an ensemble into a store')
        parser.add_argument(
            "--id", "-i", help="Desired Id for dataset", default=None)
        args, user_params = parser.parse_known_args(sys.argv[2:])
        params = {}
        index = 0
        while index < len(user_params):
            term = user_params[index]
            sp = term.split("=")
            if len(sp) > 1:
                params[sp[0]] = eval(sp[1])
                index += 1
            else:
                params[sp[0]] = eval(user_params[index+1])
                index += 2

        print("Adding ensemble to: {}".format(args.store))
        store = kosh.connect(args.store)
        store.create_ensemble(id=args.id, metadata=params)

    def add(self):
        """add a dataset to a Kosh store command"""
        parser = core_parser(
            prog="kosh add",
            description='Adds a dataset to a store and possibly an ensemble')
        parser.add_argument(
            "--ensemble", "-e", help="Id of the ensemble we want to add this to", default=None)
        parser.add_argument(
            "--id", "-i", help="Desired Id for dataset", default=None)
        args, metadata = parser.parse_known_args(sys.argv[2:])
        metadata = parse_metadata(metadata)
        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        ds = store.create(id=args.id, metadata=metadata)
        if args.ensemble is not None:
            ensemble = next(store.find_ensembles(id=args.ensemble))
            ensemble.add(ds)
            print(ensemble)
        print(ds.id)

    def remove(self):
        """Remove dataset(s) and or ensemble(s) from store command"""
        parser = core_parser(
            prog="kosh remove",
            description='Removes a dataset from store')
        parser.add_argument("--ids", "-i", help="ids of datasets/ensembles to remove",
                            nargs="*", required=True, action="append")
        parser.add_argument("--force", "-f", action="store_true",
                            help="remove without asking for confirmation")
        args = parser.parse_args(sys.argv[2:])
        datasets = []
        for i in args.ids:
            datasets += i

        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        for Id in datasets:
            if args.force:
                store.delete(Id)
            else:
                ds = store.open(Id)
                print(ds)
                answer = input(
                    "You are about the remove this dataset ({Id}). Do you want to continue? (y/N)".format(Id=Id))
                if answer.lower() in ["y", "yes"]:
                    store.delete(Id)
                else:
                    print("Skipping, will not remove {Id}".format(Id=Id))

    def features(self):
        """List features for a dataset command"""
        parser = core_parser(
            prog="kosh features",
            description='List features for (some) dataset(s)')
        parser.add_argument("--ids", "-i", help="ids of datasets to list features from",
                            nargs="*", required=True, action="append")
        args = parser.parse_args(sys.argv[2:])
        datasets = []
        for i in args.ids:
            datasets += i

        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        for Id in datasets:
            ds = store.open(Id)
            print("\nDataset: {}:\n\t {}".format(Id, ds.list_features()))

    def extract(self):
        """Extract feature from dataset"""
        parser = core_parser(
            prog="kosh extract",
            description='Extract features from a dataset')
        parser.add_argument(
            "--id", "-i", help="id of datasets to extract features from", required=True)
        parser.add_argument(
            "--features", "-f", help="features to extract", nargs="*", action="append")
        parser.add_argument(
            "--format", "-F", help="format to extract to", default="numpy")
        parser.add_argument("--dump", help="Dump to file")
        args, extract_terms = parser.parse_known_args(sys.argv[2:])
        extract_terms = parse_metadata(extract_terms)
        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        ds = store.open(args.id)
        features = []
        if args.features is not None:
            for f in args.features:
                features += f
        else:
            features = ds.list_features()

        if args.dump is not None:
            out = open(args.dump, "wb")
        for feat in features:
            data = ds.get(feat, format=args.format, **extract_terms)
            if args.dump is not None:
                try:
                    import numpy
                    numpy.save(out, data)
                except Exception:
                    print(
                        "Could not save feature {feat} to file: {args.dump}".format(feat=feat, args=args))
            else:
                print(data)

    def print(self):
        """print dataset(s)"""
        parser = core_parser(
            prog="kosh print",
            description='Print information about a dataset')
        parser.add_argument("--ids", "-i", help="ids of datasets to print",
                            nargs="*", required=True, action="append")
        args = parser.parse_args(sys.argv[2:])
        datasets = []
        for i in args.ids:
            datasets += i

        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        for Id in datasets:
            ds = store.open(Id)
            print(ds)
            print(
                "=======================================================================")

    def associate(self):
        """Associate uri with dataset command"""
        parser = core_parser(
            prog="kosh associate",
            description="Associate a (set of) files with a dataset or ensemble")
        parser.add_argument(
            "--id", "-i", help="id of datasets/ensembles to which file(s) will be associated", required=True)
        parser.add_argument("--uri", "-u", help="uri(s) to associate with dataset",
                            nargs="*", required=True, action="append")
        parser.add_argument(
            "--mime_type", "-m", help="mime type of the uri(s) same for all", required=True)
        args = parser.parse_args(sys.argv[2:])
        uris = []
        for u in args.uri:
            uris += u
        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        ds = store.open(args.id)
        for u in uris:
            ds.associate(u, mime_type=args.mime_type)

    def dissociate(self):
        """dissociate uri from dataset/ensemble command"""
        parser = core_parser(
            prog="kosh dissociate",
            description="dissociate a (set of) file(s) from a dataset")
        parser.add_argument(
            "--id", "-i", help="id of datasets/ensembles from which file(s) will be dissociated", required=True)
        parser.add_argument("--uri", "-u", help="uri(s) to dissociate from dataset",
                            nargs="*", required=True, action="append")
        args = parser.parse_args(sys.argv[2:])
        uris = []
        for u in args.uri:
            uris += u
        store = kosh.KoshStore(
            db_uri=args.store, dataset_record_type=args.dataset_record_type)
        ds = store.open(args.id)
        for u in uris:
            ds.dissociate(u)

    def mv(self):
        """mv files command"""
        if sys.platform.startswith("win"):
            raise SystemError("you cannot use kosh mv on a windows system,"
                              " please try to tar/untar or manually move the files and use reassociate")
        self._mv_cp_("mv")

    def cp(self):
        if sys.platform.startswith("win"):
            raise SystemError("you cannot use kosh cp on a windows system,"
                              " please manually cp the files and associate them")
        """cp files command"""
        self._mv_cp_("cp")

    def tar(self):
        """tar files"""
        self._tar("tar")

    def htar(self):
        """tar files using htar"""
        self._tar("htar")

    def _tar(self, tar_command, store_args="", opts=""):

        description = f"Uses `{tar_command}` to (un)tar files and the dataset"\
                        "they're associated with in selected Kosh store(s)"
        """tar files command"""
        parser = argparse.ArgumentParser(
            prog="kosh tar",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description=description,
            epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
        parser.add_argument("--stores", "--store", "-s", required=True,
                            help="Kosh store(s) to use", action="append")
        parser.add_argument("--dataset_record_type", default="dataset",
                            help="record type used by Kosh when adding datasets to Sina database")
        parser.add_argument("-f", "--file", help="tar file", required=True)
        parser.add_argument(
            "--no_absolute_path",
            action="store_true",
            help="Do not use absolute path when searching stores")
        parser.add_argument("--dataset_matching_attributes", default=["name", ],
                            help="List of attributes used to identify if two datasets are identical",
                            type=ast.literal_eval)
        parser.add_argument("--merge_strategy", help="When importing dataset, how do we handle conflict",
                            default=None, choices=["conservative", "preserve", "overwrite"])

        if store_args == "":
            args, opts = parser.parse_known_args(sys.argv[2:])
        else:
            args, _ = parser.parse_known_args(store_args)

        # Ok are we creating or extracting?
        extract = False
        create = False
        if "x" in opts[0] or "-x" in opts:
            extract = True
            if "v" not in opts[0] and "-v" not in opts:
                opts.append("-v")
        if "c" in opts[0] or "-c" in opts:
            create = True

        if "t" in opts[0] or "-t" in opts:
            raise ValueError("t (test archive) option is not supported yet")

        if create == extract:
            raise RuntimeError(
                "Could not determine if you want to create or extract the archive, aborting!")

        # Open the stores
        stores = open_stores(args.stores, args.dataset_record_type)

        if create:
            clean_json = True  # windows needs us to del manually
            # Because we cannot add the exported dataset file to a compressed archive
            # We need to figure out the list of files first
            # They should all be in opts and the tar file is not because of -f
            # option
            no_tarred_files = False
            tarred_files = get_all_files(opts)
            if tarred_files == []:
                no_tarred_files = True
                # Ok user did not pass files
                # that means we need to tar
                # all files in store
                for store in stores:
                    recs = store.get_sina_records()
                    file_type = recs.get("__kosh_store_info__")["data"]["sources_type"]["value"]
                    # Could check the file exists
                    tarred_files += [x["data"]["uri"]["value"] for x in recs.find_with_type(file_type)]

            # Prepare dictionary to hold list of datasets to export (per store)
            store_datasets = {}
            for store in stores:
                store_datasets[store.db_uri] = []

            # for each tar file let's see if it's associated to a/many
            # dataset(s) in the store
            for filename in tarred_files:
                if not args.no_absolute_path:
                    filename = os.path.abspath(filename)
                for store in stores:
                    recs = store.get_sina_records()
                    file_type = recs.get("__kosh_store_info__")["data"]["sources_type"]["value"]
                    store_ds = [x["data"]["associated"]["value"] for x in recs.find(
                        data={"uri": filename}, types=[file_type, ])]
                    for ds in store_ds:
                        store_datasets[store.db_uri] += ds

            # Ok now we need to export all the datasets to a file
            # First item is the root from where we ran the command (untar will
            # need this)
            datasets_jsons = [os.getcwd(), ]
            for store in stores:
                for Id in set(
                        store_datasets[store.db_uri]):  # loop throuh datasets but make sure only once
                    datasets_jsons.append(store.export_dataset(Id))

            # Ok let's dump this
            tmp_json = tempfile.NamedTemporaryFile(prefix="__kosh_export__",
                                                   suffix=".json",
                                                   dir=os.path.abspath(os.path.dirname(args.file)),
                                                   mode="w",
                                                   delete=False)
            json.dump(datasets_jsons, tmp_json)
            # Make sure it's all in the file before tarring it
            tmp_json.file.flush()
            if no_tarred_files:
                tarred_files = [os.path.relpath(x) for x in tarred_files]
                opts += tarred_files

            tmp_json.file.close()
            # Let's tar this!
            cmd = "{} -f {} {} {}".format(tar_command, args.file, " ".join(opts), tmp_json.name)
        else:  # ok we are extracting
            clean_json = False  # no json created
            cmd = "{} -f {} {}".format(tar_command, args.file, " ".join(opts))

        p, out, err = process_cmd(cmd)
        if sys.platform.startswith("win"):  # windows tar put message in err
            err, out = out, err
        if clean_json:
            os.remove(tmp_json.name)
        if p.returncode != 0:
            raise RuntimeError(
                "Could not run {} cmd: {}\nReceived error: {}".format(
                    tar_command, cmd, err.decode()))

        if extract:
            # ok we extracted that's nice
            # Now let's populate the stores

            # Step 1 figure out the json file that contains our datasets
            filenames = out.decode().split("\n")
            if sys.platform.startswith("win"):
                filenames = [filename.split()[1] for filename in filenames if len(filename.split()) > 1]
            if "HTAR" in filenames[0]:
                # htar used
                filenames = filenames[:-3]  # last 3 lines are nothing
                filenames = [x.split(",")[0].split()[-1].strip() for x in filenames]
            # tar removes leading slah from full path
            slashed_filenames = ["/" + x for x in filenames]
            for ifile, filename in enumerate(filenames):
                base = os.path.basename(filename)
                if base.startswith("__kosh_export__") and base.endswith(".json"):
                    break
            with open(filename) as f:
                datasets = json.load(f)

            os.remove(filename)

            # Step 2 recover the root path from where the tar was made
            # And our guessed tarrred files
            root = datasets.pop(0)
            root_filenames = [os.path.join(root, x) for x in filenames]

            # Step 3 let's put these datasets into the stores
            orphans = []
            for dataset in datasets:
                # Let's try to recover the correct path now..
                delete_them = []
                for index, associated in enumerate(dataset["records"][1:]):
                    uri = associated["data"]["uri"]["value"]
                    if uri in filenames:
                        new_uri = os.path.join(os.getcwd(), uri)
                    elif uri in slashed_filenames:  # tar removes leading /
                        new_uri = os.path.join(
                            os.getcwd(), filenames[slashed_filenames.index(uri)])
                    elif uri in root_filenames:
                        new_uri = os.path.join(
                            os.getcwd(), filenames[root_filenames.index(uri)])
                    else:
                        if not os.path.exists(uri):
                            delete_them.append(index + 1)
                        else:
                            orphans.append(uri)
                        new_uri = uri
                    associated["data"]["uri"]["value"] = new_uri

                # Yank uris that do not exists in this filesystem
                for index in delete_them[::-1]:
                    dataset["records"].pop(index)

                # Add dataset to store(s)
                for store in stores:
                    store.import_dataset(
                        dataset, args.dataset_matching_attributes,
                        merge_handler=args.merge_strategy)

            # Trying to reassociate these orphan files
            # path aliases might cause them to appear
            matches = {}
            for orphan in orphans:
                matches[orphan] = []
                # now let's try to find a possible match
                for myfile in files:  # noqa: F821
                    if len(myfile) < 2:
                        continue
                    if orphan.endswith(myfile):
                        matches[orphan].append(myfile)

            # go through the matches
            for match in matches:
                for dataset in s.find(file=match):  # noqa: F821
                    # If fast_sha changed we're hosed
                    # Trying to fix this
                    dataset.cleanup_files(clean_fastsha=True)
                    for possible in matches[match]:
                        dataset.reassociate(possible)

    def rm(self):
        """rm files command"""
        parser = argparse.ArgumentParser(
            prog="kosh rm",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="delete files and update their uris in selected Kosh store(s)",
            epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
        parser.add_argument("--stores", "--store", "-s", required=True,
                            help="Kosh store(s) to use", action="append")
        parser.add_argument("--sources", "--source", "-S", "-f", "--file", "--files",
                            help="source (files or directories) to move",
                            action="append", nargs="+", required=True)
        parser.add_argument("--dataset_record_type", default="dataset",
                            help="type used by sina db that Kosh will recognize as dataset")
        args, opts = parser.parse_known_args(sys.argv[2:])

        # Ok first step is to list files to yank
        sources = []
        for source in args.sources:
            if isinstance(source, list):
                sources += source
            else:
                sources.append(source)
        files = []
        for source in sources:
            files += glob.glob(source)

        # Ok now we need abspaths
        files = [os.path.abspath(f) for f in files]

        # Finally we need to see if there are directories in this
        del_indices = []
        for i, f in enumerate(list(files)):
            if os.path.isdir(f):
                del_indices.append(i)
                for root, dirs, filenames in os.walk(f):
                    for filename in filenames:
                        files.append(os.path.join(root, filename))
        # remove dirs
        for index in del_indices[::-1]:
            files.pop(index)

        # Ok now open the stores
        stores = open_stores(args.stores, args.dataset_record_type)

        # And finally delete the files
        for filename in files:
            try:
                os.remove(filename)
            except Exception:
                continue

            # it worked let's remove it from stores
            for store in stores:
                datasets = store.find(file=filename)
                for dataset in datasets:
                    dataset.dissociate(filename)

    def fast_sha(self):
        """print fast_sha Kosh would compute for a list of files"""
        parser = argparse.ArgumentParser(
            prog="kosh fast_sha",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="print short shas that kosh will compute for some sources",
            epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
        args, opts = parser.parse_known_args(sys.argv[2:])

        files = get_all_files(opts)
        for filename in files:
            sha = kosh.utils.compute_fast_sha(filename)
            print("{} {}".format(sha, filename))

    def reassociate(self):
        """reassociate files with datasets"""
        parser = argparse.ArgumentParser(
            prog="kosh reassociate",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="reassociate uris in selected Kosh store(s)",
            epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
        parser.add_argument("--stores", "--store", "-s", required=True,
                            help="Kosh store(s) to use", action="append")
        parser.add_argument("--new_uris", "--new_uri", "-n",
                            help="new uri(s) to use instead of old ones",
                            action="append", nargs="+", required=True)
        parser.add_argument("--original_uris", "--original_uri",
                            help="uri or sha in store that we wish to replace," +
                            " if not passed then will use short sha(s) of new_uri(s)",
                            action="append", nargs="+", default=None)
        parser.add_argument(
            "--no_absolute_path",
            action="store_true",
            help="Do not change paths to absolute path")
        parser.add_argument("--dataset_record_type", "-d", default="dataset",
                            help="type used by sina db that Kosh will recognize as dataset")

        args, opts = parser.parse_known_args(sys.argv[2:])

        new_uris = find_files_from_list(args.new_uris)

        if args.original_uris is not None:
            original_uris = find_files_from_list(args.original_uris)
            if len(original_uris) != len(new_uris):
                raise RuntimeError(
                    "you are trying to reassociate {} uris to {} new uris. Aborting!".format(
                        len(original_uris), len(new_uris)))
        else:
            original_uris = None

        for store_uri in args.stores:
            store = kosh.KoshStore(
                store_uri, dataset_record_type=args.dataset_record_type)
            for index, target in enumerate(new_uris):
                if original_uris is None:
                    store.reassociate(
                        target, absolute_path=not args.no_absolute_path)
                else:
                    store.reassociate(
                        target,
                        original_uris[index],
                        absolute_path=not args.no_absolute_path)

    def _mv_cp_(self, command, store_args=""):
        """core function to implement mv and cp"""
        if command == "mv":
            command_str = "move"
        elif command == "cp":
            command_str = "copy"
        else:
            raise ValueError("Unknown command {}".format(command))

        parser = argparse.ArgumentParser(
            prog="kosh {}".format(command),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="{} files and update their uris in selected Kosh store(s)".format(
                command_str),
            epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
        parser.add_argument("--stores", "--store", "-s", required=True,
                            help="Kosh store(s) to use", action="append")
        if command != "rm":
            parser.add_argument("--destination-stores", "--destination-store",
                                "--destination_stores", "--destination_store",
                                help="Kosh store(s) to use", action="append", default=[])
        parser.add_argument("--sources", "--source", "-S", "-f", "--file", "--files",
                            help="source (files or directories) to move",
                            action="append", nargs="+", required=True)
        parser.add_argument("--dataset_record_type", "-d", default="dataset",
                            help="type used by sina db that Kosh will recognize as dataset")
        parser.add_argument("--dataset_matching_attributes", default=["name", ],
                            help="List of attributes used to identify if two datasets are identical",
                            type=ast.literal_eval)
        if command != "rm":
            parser.add_argument("--destination",
                                help="destination (file or directory) name", required=True)
        parser.add_argument("--version", action="store_true",
                            help="print version and exit")
        parser.add_argument("--merge_strategy", help="When importing dataset, how do we handle conflict",
                            default=None, choices=["conservative", "preserve", "overwrite"])
        parser.add_argument("--mk_dirs", action="store_true",
                            help="Make destination directories if they don't exist")

        if store_args == "":
            args, opts = parser.parse_known_args(sys.argv[2:])
        else:
            args, opts = parser.parse_known_args(store_args)

        files = []
        for source in args.sources:
            if isinstance(source, list):
                files += source
            else:
                files.append(source)
        for filename in files:
            if "://" in filename:
                if len(files) > 1:
                    raise ValueError(
                        "URIS can only be moved one at a time `kosh mv --stores=[...] ur1 ur2")

        if command in ["cp", "mv"]:
            target = args.destination
        else:
            target = "/tmp"

        if command == "mv":   # Move only works on local files
            if is_remote(target) or sum([is_remote(x) for x in files]) > 0:
                raise ValueError("kosh mv only works on local files")

        sources, targets, source_dict = find_sources_and_targets(opts, files, target, args.mk_dirs)

        if command == "rm":
            targets = ["", ] * len(sources)

        # for i, source in enumerate(sources):
        #     print("Doing: {} to {}".format(source,targets[i]))

        origin_stores = open_stores(args.stores, args.dataset_record_type)

        dest_stores = open_stores(
            args.destination_stores,
            args.dataset_record_type)
        # Ok now let's prepare the actual command
        if command in ["mv", "cp"]:
            cmd = "rsync -v -r " + " ".join(opts)
        elif command in ["rm"]:
            cmd = "rm " + " ".join(opts)
        if command == "mv":
            # let's yank the source
            cmd += " --remove-source-files "
        rsync_ran = []
        for i, source in enumerate(sources):
            # Now let's run the command and see if it worked
            # But only if not ran for directory before
            if os.path.dirname(source) + "/" not in sources[:i]:
                skip_it = False
                for ran in rsync_ran:
                    if ran in source:
                        skip_it = True
                        break
                if not skip_it:
                    rsync_ran.append(source)
                    cmd_run = cmd + " {} {}".format(source, targets[i])
                    p, o, e = process_cmd(cmd_run)
                    if p.returncode != 0:  # Failed!
                        raise RuntimeError(
                            "Error running command {}, aborting!\n{}".format(
                                cmd_run, e.decode()))

            for o_store in origin_stores:
                datasets = o_store.find(file=source)

                # find_sources_and_targets() returns abs path with removed links
                # but dataset might only have local path or linked path
                datasets_search = o_store.find(file=source)  # Don't want to use generator up with if statement below
                if not list(datasets_search):
                    try:
                        source = source_dict[source]
                        datasets = o_store.find(file=source)
                    except KeyError:
                        datasets = []

                for dataset in datasets:
                    associated_uris = dataset.find(uri=source)
                    if command == "mv":
                        for associated in associated_uris:
                            associated.uri = targets[i]
                        # we also need to update the file section of our record
                        rec = dataset.get_record()
                        rec["files"][targets[i]] = rec["files"][source]  # uri needs to be the same for moved files
                        del (rec["files"][source])
                    else:  # cp
                        associated_found = False
                        for associated in associated_uris:
                            associated_found = True
                            dataset.associate(targets[i], associated.mime_type)
                        rec = dataset.get_record()

                        # Sometimes the dataset might have source in ['files'] but was never associated
                        if not associated_found:
                            rec["files"][targets[i]] = {}  # uri needs to be different for copied files

                    o_store.get_sina_records().update(rec)
                    exported = dataset.export()

                    for d_store in dest_stores:
                        d_store.import_dataset(
                                exported, args.dataset_matching_attributes,
                                merge_handler=args.merge_strategy)

            # Ok it's good let's sync the stores
            for store in origin_stores + dest_stores:
                store.sync()

        # closes stores and send them back to remote if necessary
        close_stores(origin_stores, args.stores)
        close_stores(dest_stores, args.destination_stores)

    def create_new_db(self):
        """Creates a Kosh store"""
        parser = argparse.ArgumentParser(
            prog="kosh create_new_db",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="Creates a new Kosh store",
            epilog="Kosh version {kosh.__version__}".format(kosh=kosh))
        parser.add_argument("--uri", "-u", help="path to database", required=True)
        parser.add_argument("--database", "--db", "-d", help="Database type to use as backend",
                            choices=["sql", "cass"], default="sql")
        parser.add_argument("--token", "-t", help="Token to use (for Cassandra databases)", default="")
        parser.add_argument("--keyspace", "-k", help="keyspace to use (for Cassandra databases)")
        parser.add_argument("--cluster", "-c", help="cluster to use (for Cassandra databases)")

        args = parser.parse_args(sys.argv[2:])

        kosh.create_new_db(args.uri, db=args.database,
                           token=args.token, keyspace=args.keyspace, cluster=args.cluster)

    def create(self):
        """Creates a Kosh dataset in a store"""
        parser = core_parser(
            description='Create a dataset in the store with matching metadata in form key=value')
        args, user_params = parser.parse_known_args(sys.argv[2:])

        params = {}
        index = 0
        while index < len(user_params):
            term = user_params[index]
            sp = term.split("=")
            if len(sp) > 1:
                params[sp[0]] = eval(sp[1])
                index += 1
            else:
                params[sp[0]] = eval(user_params[index+1])
                index += 2

        print("Adding ds to: {}".format(args.store))
        store = kosh.connect(args.store)
        store.create(metadata=params)


def is_remote(path):
    """Determine if a uri is located on a remote server
    First figures out if there is a ':' in the path (e.g user@host:/path)
    Then looks if there is a single @ in part preceding the first ':'
    :param path: uri/path
    :type path: str
    :return: True if path points to remote server, False otherwise
    :rtype: bool
    """
    dot = path.split(":")[0]
    at = dot.split("@")
    if len(at) == 1:
        return False
    return True


def get_realpath_and_status(source):
    """figure out if a path is:
        local or remote
        a dir or not
        if it exists or not
    works with remote paths as well.
    Return absolute path on host to directory where file exist or directory
    :param source: path to use
    :type source: str
    :return: abs path on host, is_it_remote, is_it_a_directory, does_it_exist
    :rtype: str, bool, bool, bool
    """
    is_dir_cmd = "if [ -d {} ]; then echo -e 1 ; else echo -e 0 ;  fi ;"
    realpath_cmd = "realpath {}"
    sp = source.split("@")
    if len(sp) > 1:
        is_remote = True
        user = sp[0]
        machine = sp[1].split(":")[0]
        files = ":".join("@".join(sp[1:]).split(":")[1:])
        # Find user and machine
        ssh_prefix = "ssh {}@{} ".format(user, machine)
        is_dir_cmd = ssh_prefix + "'{}'".format(is_dir_cmd)
        realpath_cmd = ssh_prefix + "'{}'".format(realpath_cmd)
    else:
        is_remote = False
        files = source
    is_dir_cmd = is_dir_cmd.format(files)
    realpath_cmd = realpath_cmd.format(files)
    if "*" in is_dir_cmd or "?" in is_dir_cmd:
        is_dir = 0
        is_file = 1
    else:
        _, o, _ = process_cmd(is_dir_cmd, use_shell=True)
        o = o.decode().split("\n")
        is_dir = int(o[0])
        cmd = is_dir_cmd.replace("-d", "-f")
        _, o, _ = process_cmd(cmd, use_shell=True)
        o = o.decode().split("\n")
        is_file = int(o[0])
    if is_dir == 0 and is_file == 0:
        # Ok it does not exist
        exists = False
    else:
        exists = True
    _, realpath, _ = process_cmd(realpath_cmd.format(files), use_shell=True)
    return realpath.decode().strip().split("\n")[0], is_remote, is_dir, exists


def find_depth(path):
    """Given a path returns how level of directories this is in
    :param path: path to scan
    :type path: str
    :return: number of directories in which the file is in this path
    :rtype: int
    """
    depth = 0
    tmp = os.path.split(path)
    while tmp[-1] != "":
        depth += 1
        tmp = os.path.split(tmp[0])
    return depth - 1


def find_sources_and_targets(options, sources, target, mk_dirs=False):
    """Given a list of sources (files, dir, patterns) and a target destination,
    runs 'rsync' between these to obtain the list of files being touched
    :param options: option to send to rsync
    :type options: list
    :param sources: list of sources files, dirs or patterns
    :type sources: list
    :param target: target file or directory
    :type target: str
    :param mk_dirs: Make destination directories if they don't exist
    :type mk_dirs: bool, optional
    :return: List of sources and there matching path after cp/mv
    :rtype: list, list, dict
    """

    if mk_dirs and not os.path.islink(target):
        split = os.path.split(target)
        root_ext = os.path.splitext(target)
        if root_ext[1] == '':
            os.makedirs(target, exist_ok=True)
        else:
            os.makedirs(split[0], exist_ok=True)

    target_realpath, is_target_remote, is_target_dir, target_exists = get_realpath_and_status(
        target)
    source_uris = []
    target_uris = []
    source_dict = {}
    for source in sources:
        source_realpath, is_source_remote, is_source_dir, source_exists = get_realpath_and_status(
            source)
        source_dict[source_realpath] = source

        if is_source_dir and source[-1] == '/':
            source = source[:-1]

        if len(sources) > 1 and not is_source_dir and not target_exists:
            raise ValueError(
                "Destination ({}) does not exist and you're trying to send multiple sources to it".format(target))
        if not source_exists:
            raise RuntimeError("Source {} does not exists".format(source))

        if os.path.islink(source):
            source = source_realpath

        cmd = "rsync -v --dry-run -r" + " ".join(options)
        cmd += " " + source + " " + target
        p, o, e = process_cmd(cmd, use_shell=True)
        rsync_dryrun_out_lines = o.decode().split("\n")
        found_a_dir_to_rsync = False
        for i, ln in enumerate(rsync_dryrun_out_lines):
            if i == 0:
                depth = find_depth(ln) + 1
                counter = 0
                non_base = []
                abs_path = source_realpath
                while counter != depth:
                    non_base.insert(0, os.path.basename(abs_path))
                    abs_path = os.path.dirname(abs_path)
                    counter += 1
                non_base = os.path.join(*non_base)
            if len(ln.strip()) == 0:
                break
            elif len(shlex.split(ln.strip())) == 1:
                if ln.strip()[-1] == "/":
                    if found_a_dir_to_rsync is False:
                        source_uris.append(os.path.join(
                            abs_path, ln.strip()[:-1]))
                        found_a_dir_to_rsync = True
                    else:
                        continue
                else:
                    source_uris.append(os.path.join(abs_path, ln.strip()))
                if target_exists:
                    if ln.strip()[-1] == "/":
                        # it's a dir
                        target_uris.append(target_realpath)
                    else:
                        if mk_dirs and os.path.split(target)[1] == ln.strip():
                            target_uris.append(target_realpath)
                        else:
                            if is_target_dir:
                                target_uris.append(os.path.join(
                                    target_realpath, ln.strip()))
                            else:
                                target_uris.append(target_realpath)
                else:
                    if ln.strip()[-1] == "/" or not is_source_dir:
                        target_uris.append(target_realpath)
                    else:
                        target_uris.append(os.path.join(
                            target_realpath, ln.strip()))
    return source_uris, target_uris, source_dict


if __name__ == '__main__':
    KoshCmd()
